# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import jax.numpy as jnp
import numpy as np
from jax import Array

import sepes._src.backend.arraylib as arraylib

arraylib.tobytes.register(Array, lambda x: np.array(x).tobytes())
arraylib.where.register(Array, jnp.where)
arraylib.nbytes.register(Array, lambda x: x.nbytes)
arraylib.shape.register(Array, jnp.shape)
arraylib.dtype.register(Array, lambda x: x.dtype)
arraylib.min.register(Array, jnp.min)
arraylib.max.register(Array, jnp.max)
arraylib.mean.register(Array, jnp.mean)
arraylib.std.register(Array, jnp.std)
arraylib.all.register(Array, jnp.all)
arraylib.array_equal.register(Array, np.array_equal)  # NOTE: not traceable
arraylib.is_floating.register(Array, lambda x: jnp.issubdtype(x.dtype, jnp.floating))
arraylib.is_integer.register(Array, lambda x: jnp.issubdtype(x.dtype, jnp.integer))
arraylib.is_inexact.register(Array, lambda x: jnp.issubdtype(x.dtype, jnp.inexact))
arraylib.is_bool.register(Array, lambda x: jnp.issubdtype(x.dtype, jnp.bool_))
arraylib.ndarrays.append(Array)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constructor code generation from type annotations."""

# this modules contains functionality to turn type hints into a constructor
# similar to `dataclasses.dataclass`/`attrs`
# however, notable differences are:
# - Fields are not tied to the class decorator. i.e. `Field` can be used without `autoinit`.
# - Fields enable running function callbacks on the field values during setting/getting.
#   using descriptors and emit a descriptive error message in case of an error.
# - Marking fields as positional only, keyword only, variable positional,...
# - Does not allow mutable defaults.
# - Registering additional types to be excluded from `autoinit`. e.g. raise an error.
# - Only code generation is supported is done. other functionality like `__repr__`,
#   `__eq__`, `__hash__`,are not supported.

# one design choice is that `autoinit` and `Field` are not tightly coupled.
# Field` can be used without `autoinit` as a descriptor to apply functions on
# the field values during initialization. Moreover, `TreeClass` is not coupled with
# `autoinit` or `Field` and can be used without them. this simplifies the code
# by separating the functionality.

from __future__ import annotations

import functools as ft
import sys
from collections import defaultdict
from collections.abc import Callable, MutableMapping, MutableSequence, MutableSet
from typing import Any, Literal, Sequence, TypeVar, get_args
from warnings import warn
from weakref import WeakSet
from typing_extensions import dataclass_transform


T = TypeVar("T")
PyTree = Any
EllipsisType = type(Ellipsis)
KindType = Literal["POS_ONLY", "POS_OR_KW", "VAR_POS", "KW_ONLY", "VAR_KW", "CLASS_VAR"]
arg_kinds: tuple[str, ...] = get_args(KindType)
EXCLUDED_FIELD_NAMES: set[str] = {"self", "__post_init__", "__annotations__"}
_autoinit_registry: WeakSet[type] = WeakSet()


class Null:
    __slots__ = []
    __repr__ = lambda _: "NULL"
    __bool__ = lambda _: False


NULL = Null()


def generate_field_doc(field: Field) -> str:
    out: list[str] = ["Field Information:"]
    out += [f"\tName:\t\t``{field.name}``"]
    out += [f"\tDefault:\t``{field.default}``"] if field.default is not NULL else []
    out += [f"Description:\n\t{field.doc}"] if field.doc else []

    if field.on_setattr or field.on_getattr:
        out += ["Callbacks:"]

    if field.on_setattr:
        out += ["\t- On setting attribute:\n"]
        out += [f"\t\t- ``{func}``" for func in field.on_setattr]

    if field.on_getattr:
        out += ["\t- On getting attribute:\n"]
        out += [f"\t\t- ``{func}``" for func in field.on_getattr]

    return "\n".join(out)


def slots(klass) -> tuple[str, ...]:
    return getattr(klass, "__slots__", ())


def pipe(funcs: Sequence[Callable[[Any], Any]], name: str | None, value: Any):
    """Apply a sequence of functions on the field value."""
    for func in funcs:
        # for a given sequence of unary functions, apply them on the field value
        # and return the result. if an error is raised, emit a descriptive error
        try:
            value = func(value)
        except Exception as e:
            # emit a *descriptive* error message with the name of the attribute
            # associated with the field and the name of the function that raised
            # the error.
            cname = getattr(func, "__name__", func)
            raise type(e)(f"On applying {cname} for field=`{name}`:\n{e}")
    return value


class Field:
    """Field descriptor placeholder"""

    __slots__ = [
        "name",
        "type",
        "default",
        "init",
        "repr",
        "kind",
        "metadata",
        "on_setattr",
        "on_getattr",
        "alias",
        "doc",
    ]

    def __init__(
        self,
        *,
        name: str | Null = NULL,
        type: type | Null = NULL,
        default: Any = NULL,
        init: bool = True,
        repr: bool = True,
        kind: KindType = "POS_OR_KW",
        metadata: dict[str, Any] | None = None,
        on_setattr: Sequence[Callable[[Any], Any]] = (),
        on_getattr: Sequence[Callable[[Any], Any]] = (),
        alias: str | None = None,
        doc: str = "",
    ):
        self.name = name
        self.type = type
        self.default = default
        self.init = init
        self.repr = repr
        self.kind = kind
        self.metadata = metadata
        self.on_setattr = on_setattr
        self.on_getattr = on_getattr
        self.alias = alias
        self.doc = doc

    def replace(self, **kwargs) -> Field:
        """Replace the field attributes."""
        # define a `replace` method similar to `dataclasses.replace` or namedtuple
        # to allow the user to replace the field attributes.
        return type(self)(**{k: kwargs.get(k, getattr(self, k)) for k in slots(Field)})

    def pipe_on_setattr(self, value: Any) -> Any:
        """Apply a sequence of functions on the field value during setting."""
        return pipe(self.on_setattr, self.name, value)

    def pipe_on_getattr(self, value: Any) -> Any:
        """Apply a sequence of functions on the field value during getting."""
        return pipe(self.on_getattr, self.name, value)

    def __set_name__(self, owner, name: str) -> None:
        """Set the field name."""
        # set the name of the field to the attribute name in the class
        # and the type to the type hint of the attribute if it exists
        self.name = name
        # in case the user uses `field` as a descriptor without annotating the class
        if "__annotations__" in (variables := vars(owner)):
            # set the type to the type hint of the attribute if it exists
            self.type = variables.get(name, NULL)

    @property
    def __doc__(self) -> str:
        """Return the field documentation."""
        return generate_field_doc(field=self)

    def __get__(self: T, instance, _) -> T | Any:
        """Return the field value."""
        if instance is None:
            return self
        return self.pipe_on_getattr(vars(instance)[self.name])

    def __set__(self: T, instance, value) -> None:
        """Set the field value."""
        vars(instance)[self.name] = self.pipe_on_setattr(value)

    def __delete__(self: T, instance) -> None:
        """Delete the field value."""
        del vars(instance)[self.name]


def field(
    *,
    default: Any = NULL,
    init: bool = True,
    repr: bool = True,
    kind: KindType = "POS_OR_KW",
    metadata: dict[str, Any] | None = None,  # type: ignore
    on_setattr: Sequence[Any] = (),
    on_getattr: Sequence[Any] = (),
    alias: str | None = None,
    doc: str = "",
) -> Field:
    """Field placeholder for type hinted attributes.

    Args:
        default: The default value of the field.
        init: Whether the field is included in the object's ``__init__`` function.
        repr: Whether the field is included in the object's ``__repr__`` function.
        kind: Argument kind used in the constructor sythesis with :func:`autoinit`,

            - ``POS_ONLY``: positional only argument (e.g. ``x`` in ``def f(x, /):``)
            - ``VAR_POS``: variable positional argument (e.g. ``*x`` in ``def f(*x):``)
            - ``POS_OR_KW``: positional or keyword argument (e.g. ``x`` in ``def f(x):``)
            - ``KW_ONLY``: keyword only argument (e.g. ``x`` in ``def f(*, x):``)
            - ``VAR_KW``: variable keyword argument (e.g. ``**x`` in ``def f(**x):``)
            - ``CLASS_VAR``: Non-constructor class variable (e.g. ``x`` in ``class C: x = 1``)

        metadata: A mapping of user-defined data for the field.
        on_setattr: A sequence of functions to called on ``__setattr__``.
        on_getattr: A sequence of functions to called on ``__getattr__``.
        alias: An a alias for the field name in the constructor. e.g ``name=x``,
            ``alias=y`` will allow ``obj = Class(y=1)`` to be equivalent to
            ``obj = Class(x=1)``.
        doc: extra documentation for the :func:.`field` .the complete documentation
            of the field includes the field name, the field doc, and the
            default value, and function callbacks applied on the field value.
            Mainly used for documenting the field callbacks.

            .. code-block:: python

                >>> import sepes as sp
                >>> @sp.autoinit
                ... class Tree:
                ...     leaf: int = sp.field(
                ...         default=1,
                ...         doc="Leaf node of the tree.",
                ...         on_setattr=[lambda x: x],
                ... )
                >>> print(Tree.leaf.__doc__)  # doctest: +SKIP
                Field Information:
                        Name:           ``leaf``
                        Default:        ``1``
                Description:
                        Leaf node of the tree.
                Callbacks:
                        - On setting attribute:

                                - ``<function Tree.<lambda> at 0x11c53dc60>``

    Example:
        Type and range validation using :attr:`on_setattr`:

        >>> import sepes as sp
        >>> @sp.autoinit
        ... class IsInstance(sp.TreeClass):
        ...    klass: type
        ...    def __call__(self, x):
        ...        assert isinstance(x, self.klass)
        ...        return x
        <BLANKLINE>
        >>> @sp.autoinit
        ... class Range(sp.TreeClass):
        ...    start: int|float = float("-inf")
        ...    stop: int|float = float("inf")
        ...    def __call__(self, x):
        ...        assert self.start <= x <= self.stop
        ...        return x
        <BLANKLINE>
        >>> @sp.autoinit
        ... class Employee(sp.TreeClass):
        ...    # assert employee ``name`` is str
        ...    name: str = sp.field(on_setattr=[IsInstance(str)])
        ...    # use callback compostion to assert employee ``age`` is int and positive
        ...    age: int = sp.field(on_setattr=[IsInstance(int), Range(1)])
        >>> employee = Employee(name="Asem", age=10)
        >>> print(employee)
        Employee(name=Asem, age=10)

    Example:
        Private attribute using :attr:`alias`:

        >>> import sepes as sp
        >>> @sp.autoinit
        ... class Employee(sp.TreeClass):
        ...     # `alias` is the name used in the constructor
        ...    _name: str = sp.field(alias="name")
        >>> employee = Employee(name="Asem")  # use `name` in the constructor
        >>> print(employee)  # `_name` is the private attribute name
        Employee(_name=Asem)

    Example:
        Buffer creation using :attr:`on_getattr`:

        >>> import sepes as sp
        >>> import jax
        >>> import jax.numpy as jnp
        >>> @sp.autoinit
        ... class Tree(sp.TreeClass):
        ...     buffer: jax.Array = sp.field(on_getattr=[jax.lax.stop_gradient])
        >>> tree = Tree(buffer=jnp.array((1.0, 2.0)))
        >>> def sum_buffer(tree):
        ...     return tree.buffer.sum()
        >>> print(jax.grad(sum_buffer)(tree))  # no gradient on `buffer`
        Tree(buffer=[0. 0.])

    Example:
        Parameterization using :attr:`on_getattr`:

        >>> import sepes as sp
        >>> import jax
        >>> import jax.numpy as jnp
        >>> def symmetric(array: jax.Array) -> jax.Array:
        ...    triangle = jnp.triu(array)  # upper triangle
        ...    return triangle + triangle.transpose(-1, -2)
        >>> @sp.autoinit
        ... class Tree(sp.TreeClass):
        ...    symmetric_matrix: jax.Array = sp.field(on_getattr=[symmetric])
        >>> tree = Tree(symmetric_matrix=jnp.arange(9).reshape(3, 3))
        >>> print(tree.symmetric_matrix)
        [[ 0  1  2]
         [ 1  8  5]
         [ 2  5 16]]

    Note:
        - :func:`field` is commonly used to annotate the class attributes to be
          used by the :func:`autoinit` decorator to generate the ``__init__``
          method similar to ``dataclasses.dataclass``.

        - :func:`field` can be used without the :func:`autoinit` as a descriptor
          to apply functions on the field values during initialization using
          the ``on_setattr`` / ``on_getattr`` argument.

            >>> import sepes as sp
            >>> def print_and_return(x):
            ...    print(f"Setting {x}")
            ...    return x
            >>> class Tree:
            ...    # `a` must be defined as a class attribute for the descriptor to work
            ...    a: int = sp.field(on_setattr=[print_and_return])
            ...    def __init__(self, a):
            ...        self.a = a
            >>> tree = Tree(1)
            Setting 1
    """
    if not isinstance(alias, (str, type(None))):
        raise TypeError(f"Non-string {alias=} argument provided to `field`")

    if not isinstance(metadata, (dict, type(None))):
        raise TypeError(f"Non-dict {metadata=} argument provided to `field`")

    if kind not in arg_kinds:
        raise ValueError(f"{kind=} not in {arg_kinds}")

    if not isinstance(on_setattr, Sequence):
        raise TypeError(f"Non-sequence {on_setattr=} argument provided to `field`")

    if not isinstance(on_getattr, Sequence):
        raise TypeError(f"Non-sequence {on_getattr=} argument provided to `field`")

    if not isinstance(init, bool):
        raise TypeError(f"Non-bool {init=} argument provided to `field`")

    for func in on_setattr:
        if not isinstance(func, Callable):  # type: ignore
            raise TypeError(f"Non-callable {func=} provided to `field` on_setattr")

    for func in on_getattr:
        if not isinstance(func, Callable):
            raise TypeError(f"Non-callable {func=} provided to `field` on_getattr")

    return Field(
        default=default,
        init=init,
        repr=repr,
        kind=kind,
        metadata=metadata,  # type: ignore
        on_setattr=on_setattr,
        on_getattr=on_getattr,
        alias=alias,
        doc=doc,
    )


def build_field_map(klass: type) -> dict[str, Field]:
    field_map: dict[KindType, Field] = dict()

    if klass is object:
        return dict(field_map)

    for base in reversed(klass.__mro__[1:]):
        field_map.update(build_field_map(base))

    if (hint_map := vars(klass).get("__annotations__", NULL)) is NULL:
        # not annotated
        return dict(field_map)

    if EXCLUDED_FIELD_NAMES.intersection(hint_map):
        raise ValueError(f"`Field` in {EXCLUDED_FIELD_NAMES=}")

    for key, hint in hint_map.items():
        if isinstance(value := vars(klass).get(key, NULL), Field):
            # case: `x: Any = field(default=1)`
            field_map[key] = value.replace(name=key, type=hint)

    return field_map


def fields(x: Any) -> tuple[Field, ...]:
    """Returns a tuple of ``Field`` objects for the given instance or class.

    ``Field`` objects are generated from the class type hints and contains
    the information about the field information.if the user uses
    the ``sepes.field`` to annotate.

    Note:
        - If the class is not annotated, an empty tuple is returned.
        - The ``Field`` generation is cached for class and its bases.
    """
    return tuple(build_field_map(x if isinstance(x, type) else type(x)).values())


def convert_hints_to_fields(klass: type[T]) -> type[T]:
    # convert klass hints to `Field` objects for the **current** class
    if (hint_map := vars(klass).get("__annotations__", NULL)) is NULL:
        # in case no type hints are provided, return the class as is
        return klass

    for key, hint in hint_map.items():
        if isinstance(value := vars(klass).get(key, NULL), Field):
            if value.kind == "CLASS_VAR":
                setattr(klass, key, value.default)
            continue
        # no need to convert `Field` annotated attributes again
        setattr(klass, key, Field(default=value, type=hint, name=key))
    return klass


def check_excluded_types(field_map: dict[KindType, Field]) -> dict[KindType, Field]:
    # check if the user uses excluded types in `autoinit`
    # like mutable types
    for key in field_map:
        excluded_type_dispatcher(field_map[key].default)
    return field_map


def check_duplicate_var_kind(field_map: dict[KindType, Field]) -> dict[KindType, Field]:
    # check for duplicate `VAR_POS` and `VAR_KW` arguments
    seen: set[Literal["VAR_POS", "VAR_KW"]] = set()
    for field in field_map.values():
        if field.kind in ("VAR_POS", "VAR_KW"):
            if field.kind in seen:
                raise TypeError(f"Duplicate {field.kind=} for {field.name=}")
            seen.add(field.kind)
    return field_map


def check_order_of_args(field_map: dict[KindType, Field]) -> dict[KindType, Field]:
    # check if the order of the arguments is valid
    # otherwise raise a warning to the user to acknowledge the change
    # in the order of the arguments
    # for reference `dataclasses.dataclass` does not warn the user before
    # reordering the arguments
    seen: list[KindType] = []
    for key in field_map:
        if field_map[key].kind == "CLASS_VAR":
            continue
        seen += [field_map[key].kind]
        if len(seen) > 1 and arg_kinds.index(seen[-2]) > arg_kinds.index(seen[-1]):
            warn(f"Kind order {seen} order != {arg_kinds} and will be reordered.")
    return field_map


def build_init_method(klass: type[T]) -> type[T]:
    field_map: dict[KindType, Field] = build_field_map(klass)
    field_map = check_excluded_types(field_map)
    field_map = check_duplicate_var_kind(field_map)
    field_map = check_order_of_args(field_map)

    hints = {"return": None}  # annotations

    body: list[str] = []
    head: list[str] = ["self"]
    heads: dict[KindType, list[str]] = defaultdict(list)

    for field in field_map.values():
        if field.kind == "CLASS_VAR":
            # skip class variables from init synthesis
            # e.g. class A: x = field(default=1, kind="CLASS_VAR")
            continue

        if field.init:
            # add to field to head and body
            hints[field.name] = field.type
            # how to name the field in the constructor
            alias = field.alias or field.name
            body += [f"self.{field.name}={alias}"]

            if field.default is NULL:
                # e.g. def __init__(.., x)
                heads[field.kind] += [alias]
            else:
                # e.g def __init__(.., x=value) but
                # pass reference to the default value
                heads[field.kind] += [f"{alias}=refmap['{field.name}'].default"]

        else:
            if field.default is not NULL:
                # case for fields with `init=False` and no default value
                # usaully declared in __post_init__
                body += [f"self.{field.name}=refmap['{field.name}'].default"]

    # add pass in case all fields are not included in the constructor
    # i.e. `init=False` for all fields
    body += [f"self.{key}()"] if (key := "__post_init__") in vars(klass) else ["pass"]

    # organize the arguments order:
    # (POS_ONLY, POS_OR_KW, VAR_POS, KW_ONLY, VAR_KW)
    head += (heads["POS_ONLY"] + ["/"]) if heads["POS_ONLY"] else []
    head += heads["POS_OR_KW"]
    head += ["*" + "".join(heads["VAR_POS"])] if heads["VAR_POS"] else []
    # case for ...(*a, b) and ...(a, *, b)
    head += ["*"] if (heads["KW_ONLY"] and not heads["VAR_POS"]) else []
    head += heads["KW_ONLY"]
    head += ["**" + "".join(heads["VAR_KW"])] if heads["VAR_KW"] else []

    # generate the code for the method
    code = "def closure(refmap):\n"
    code += f"\tdef __init__({','.join(head)}):"
    field_map["__annotations__"] = hints

    code += f"\n\t\t{';'.join(body)}"
    code += f"\n\t__init__.__qualname__ = '{klass.__qualname__}.__init__'"
    code += "\n\t__init__.__annotations__ = refmap['__annotations__']"
    code += "\n\treturn __init__"

    # execute the code in the class namespace to generate the method
    exec(code, vars(sys.modules[klass.__module__]), namespace := dict())
    method = namespace["closure"](field_map)
    # add the method to the class
    setattr(klass, "__init__", method)
    # mark the class as transformed
    return klass


@dataclass_transform(field_specifiers=(Field, field))
def autoinit(klass: type[T]) -> type[T]:
    """A class decorator that generates the ``__init__`` method from type hints.

    Using the ``autoinit`` decorator, the user can define the class attributes
    using type hints and the ``__init__`` method will be generated automatically

    >>> import sepes as sp
    >>> @sp.autoinit
    ... class Tree:
    ...     x: int
    ...     y: int

    Is equivalent to:

    >>> class Tree:
    ...     def __init__(self, x: int, y: int):
    ...         self.x = x
    ...         self.y = y

    Example:
        >>> import sepes as sp
        >>> import inspect
        >>> @sp.autoinit
        ... class Tree:
        ...     x: int
        ...     y: int
        >>> inspect.signature(Tree.__init__)
        <Signature (self, x: int, y: int) -> None>
        >>> tree = Tree(1, 2)
        >>> tree.x, tree.y
        (1, 2)

    Example:
        Define fields with different argument kinds

        >>> import sepes as sp
        >>> import inspect
        >>> @sp.autoinit
        ... class Tree:
        ...     kw_only_field: int = sp.field(default=1, kind="KW_ONLY")
        ...     pos_only_field: int = sp.field(default=2, kind="POS_ONLY")
        >>> inspect.signature(Tree.__init__)
        <Signature (self, pos_only_field: int = 2, /, *, kw_only_field: int = 1) -> None>

    Example:
        Define a converter to apply ``abs`` on the field value

        >>> @sp.autoinit
        ... class Tree:
        ...     a:int = sp.field(on_setattr=[abs])
        >>> Tree(a=-1).a
        1

    .. warning::
        The ``autoinit`` decorator will raise ``TypeError`` if the user defines
        ``__init__`` method in the decorated class.

    Note:
        - In case of inheritance, the ``__init__`` method is generated from the
          the type hints of the current class and any base classes that
          are decorated with ``autoinit``.

        >>> import sepes as sp
        >>> import inspect
        >>> @sp.autoinit
        ... class Base:
        ...     x: int
        >>> @sp.autoinit
        ... class Derived(Base):
        ...     y: int
        >>> obj = Derived(x=1, y=2)
        >>> inspect.signature(obj.__init__)
        <Signature (x: int, y: int) -> None>

        - Base classes that are not decorated with ``autoinit`` are ignored during
          synthesis of the ``__init__`` method.

        >>> import sepes as sp
        >>> import inspect
        >>> class Base:
        ...     x: int
        >>> @sp.autoinit
        ... class Derived(Base):
        ...     y: int
        >>> obj = Derived(y=2)
        >>> inspect.signature(obj.__init__)
        <Signature (y: int) -> None>

    Note:
        Use ``autoinit`` instead of ``dataclasses.dataclass`` if you want to
        use ``jax.Array`` as a field default value. As ``dataclasses.dataclass``
        will incorrectly raise an error starting from python 3.11 complaining
        that ``jax.Array`` is not immutable.

    Note:
        By default ``autoinit`` will raise an error if the user uses mutable defaults.
        To register an additional type to be excluded from ``autoinit``, use
        :func:`autoinit.register_excluded_type`, with an optional ``reason``
        for excluding the type.

        >>> import sepes as sp
        >>> class T:
        ...     pass
        >>> sp.autoinit.register_excluded_type(T, reason="not allowed")
        >>> @sp.autoinit
        ... class Tree:
        ...     x: T = sp.field(default=T())  # doctest: +SKIP
        Traceback (most recent call last):
            ...
    """
    if klass in _autoinit_registry:
        # autoinit(autoinit(klass)) == autoinit(klass)
        # idempotent decorator to avoid redefining the class
        return klass

    if "__init__" in vars(klass):
        # if the class already has a user-defined __init__ method
        # then raise an error to avoid confusing the user
        raise TypeError(f"autoinit({klass.__name__}) with defined `__init__`.")

    # first convert the current class hints to fields
    # then build the __init__ method from the fields of the current class
    # and any base classes that are decorated with `autoinit`
    klass = build_init_method(convert_hints_to_fields(klass))
    # add the class to the registry to avoid redefining the class
    _autoinit_registry.add(klass)
    return klass


excluded_type_dispatcher = ft.singledispatch(lambda _: None)


def register_excluded_type(klass: type, reason: str | None = None) -> None:
    """Exclude a type from being used in the ``autoinit`` decorator.

    Args:
        klass: The type to be excluded.
        reason: The reason for excluding the type.
    """
    reason = f" {reason=}" if reason is not None else ""

    @excluded_type_dispatcher.register(klass)
    def _(value) -> None:
        raise TypeError(f"{value=} is excluded from `autoinit`.{reason}")


autoinit.register_excluded_type = register_excluded_type
autoinit.register_excluded_type(MutableMapping, reason="mutable type")
autoinit.register_excluded_type(MutableSequence, reason="mutable type")
autoinit.register_excluded_type(MutableSet, reason="mutable type")



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Backend tools for sepes."""

from __future__ import annotations

import functools as ft
from typing import Callable, NamedTuple


class NoImplError(NamedTuple):
    op: Callable

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(f"No implementation for {self.op}"
                                  f" with {args=} {kwargs=}")


tobytes = ft.singledispatch(NoImplError("tobytes"))
where = ft.singledispatch(NoImplError("where"))
nbytes = ft.singledispatch(NoImplError("nbytes"))
shape = ft.singledispatch(NoImplError("shape"))
dtype = ft.singledispatch(NoImplError("dtype"))
min = ft.singledispatch(NoImplError("min"))
max = ft.singledispatch(NoImplError("max"))
mean = ft.singledispatch(NoImplError("mean"))
std = ft.singledispatch(NoImplError("std"))
all = ft.singledispatch(NoImplError("all"))
array_equal = ft.singledispatch(NoImplError("array_equal"))
is_floating = ft.singledispatch(NoImplError("is_floating"))
is_integer = ft.singledispatch(NoImplError("is_integer"))
is_inexact = ft.singledispatch(NoImplError("is_inexact"))
is_bool = ft.singledispatch(NoImplError("is_bool"))
ndarrays: list[type] = []



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to work with non-inexact type tree leaves across function transformations."""

from __future__ import annotations

import functools as ft
import hashlib
from typing import Any, Callable, NamedTuple, TypeVar, Union

import sepes
import sepes._src.backend.arraylib as arraylib
from sepes._src.backend import is_package_avaiable
from sepes._src.tree_pprint import tree_repr, tree_str, tree_summary
from sepes._src.tree_util import Static, is_tree_equal, tree_copy, tree_hash

T = TypeVar("T")
MaskType = Union[T, Callable[[Any], bool]]


class _MaskedError(NamedTuple):
    opname: str

    def __call__(self, *a, **k):
        raise NotImplementedError(
            f"Cannot apply `{self.opname}` operation on a masked object "
            f"{', '.join(map(str, a))} "
            f"{', '.join(k + '=' + str(v) for k, v in k.items())}.\n"
            "Unmask the object first using `tree_unmask`"
        )


class _MaskBase(Static[T]):
    # the objective of this class is to wrap a pytree node with a custom wrapper
    # that yields no leaves when flattened. This is useful to avoid updating
    # the node by effectivly *hiding it* from function transformations that operates
    # on flattened pytrees.
    __slots__ = ["__wrapped__"]
    __wrapped__: T

    def __init__(self, node: T) -> None:
        object.__setattr__(self, "__wrapped__", node)

    def __setattr__(self, _, __) -> None:
        raise AttributeError("Cannot assign to frozen instance.")

    def __delattr__(self, _: str) -> None:
        raise AttributeError("Cannot delete from frozen instance.")

    def __repr__(self) -> str:
        return "#" + tree_repr(self.__wrapped__)

    def __str__(self) -> str:
        return "#" + tree_str(self.__wrapped__)

    def __copy__(self) -> _MaskBase[T]:
        return type(self)(tree_copy(self.__wrapped__))

    # raise helpful error message when trying to interact with frozen object
    __add__ = __radd__ = __iadd__ = _MaskedError("+")
    __sub__ = __rsub__ = __isub__ = _MaskedError("-")
    __mul__ = __rmul__ = __imul__ = _MaskedError("*")
    __matmul__ = __rmatmul__ = __imatmul__ = _MaskedError("@")
    __truediv__ = __rtruediv__ = __itruediv__ = _MaskedError("/")
    __floordiv__ = __rfloordiv__ = __ifloordiv__ = _MaskedError("//")
    __mod__ = __rmod__ = __imod__ = _MaskedError("%")
    __pow__ = __rpow__ = __ipow__ = _MaskedError("**")
    __lshift__ = __rlshift__ = __ilshift__ = _MaskedError("<<")
    __rshift__ = __rrshift__ = __irshift__ = _MaskedError(">>")
    __and__ = __rand__ = __iand__ = _MaskedError("and")
    __xor__ = __rxor__ = __ixor__ = _MaskedError("")
    __or__ = __ror__ = __ior__ = _MaskedError("or")
    __neg__ = __pos__ = __abs__ = __invert__ = _MaskedError("unary")
    __lt__ = __le__ = __gt__ = __ge__ = _MaskedError("comparison")
    __call__ = _MaskedError("__call__")


@tree_summary.def_type(_MaskBase)
def _(node) -> str:
    return f"#{tree_summary.type_dispatcher(node.__wrapped__)}"


class _MaskedHashable(_MaskBase):
    def __hash__(self) -> int:
        return tree_hash(self.__wrapped__)

    def __eq__(self, rhs: Any) -> bool:
        if not isinstance(rhs, _MaskedHashable):
            return False
        return is_tree_equal(self.__wrapped__, rhs.__wrapped__)


class _MaskedArray(_MaskBase):
    # wrap arrays with a custom wrapper that implements hash and equality
    # using the wrapped array's bytes representation and sha256 hash function
    # this is useful to select some array to hold without updating in the process
    # of training a model.
    def __hash__(self) -> int:
        bytes = arraylib.tobytes(self.__wrapped__)
        return int(hashlib.sha256(bytes).hexdigest(), 16)

    def __eq__(self, other) -> bool:
        if not isinstance(other, _MaskedArray):
            return False
        lhs, rhs = self.__wrapped__, other.__wrapped__
        # fast path to avoid calling `all` on large arrays
        if arraylib.shape(lhs) != arraylib.shape(rhs):
            return False
        if arraylib.dtype(lhs) != arraylib.dtype(rhs):
            return False
        return arraylib.array_equal(lhs, rhs)


def mask(value: T) -> _MaskBase[T]:
    # dispatching is used to customize the type of the wrapper based on the type
    # of the value. For instance, hashable values dont need custom hash and
    # equality implementations, so they are wrapped with a simpler wrapper.
    # this approach avoids type logic in the wrapper equality and hash methods,
    # thus effectively improving performance of the wrapper.
    return mask.type_dispatcher(value)


mask.type_dispatcher = ft.singledispatch(_MaskedHashable)
mask.def_type = mask.type_dispatcher.register


for ndarray in arraylib.ndarrays:

    @mask.def_type(ndarray)
    def mask_array(value: T) -> _MaskedArray[T]:
        # wrap arrays with a custom wrapper that implements hash and equality
        # arrays can be hashed by converting them to bytes and hashing the bytes
        return _MaskedArray(value)


@mask.def_type(_MaskBase)
def _(value: _MaskBase[T]) -> _MaskBase[T]:
    # idempotent mask operation, meaning that mask(mask(x)) == mask(x)
    # this is useful to avoid recursive unwrapping of frozen values, plus its
    # meaningless to mask a frozen value.
    return value


def is_masked(value: Any) -> bool:
    """Returns True if the value is a frozen wrapper."""
    return isinstance(value, _MaskBase)


def unmask(value: T) -> T:
    return unmask.type_dispatcher(value)


unmask.type_dispatcher = ft.singledispatch(lambda x: x)
unmask.def_type = unmask.type_dispatcher.register


@unmask.def_type(_MaskBase)
def _(value: _MaskBase[T]) -> T:
    return getattr(value, "__wrapped__")


def is_nondiff(value: Any) -> bool:
    return is_nondiff.type_dispatcher(value)


is_nondiff.type_dispatcher = ft.singledispatch(lambda _: True)
is_nondiff.def_type = is_nondiff.type_dispatcher.register


for ndarray in arraylib.ndarrays:

    @is_nondiff.def_type(ndarray)
    def is_nondiff_array(value) -> bool:
        # return True if the node is non-inexact type, otherwise False
        if arraylib.is_inexact(value):
            return False
        return True


@is_nondiff.def_type(float)
@is_nondiff.def_type(complex)
def _(_: float | complex) -> bool:
    return False


def _tree_mask_map(
    tree: T,
    cond: Callable[[Any], bool],
    func: type | Callable[[Any], Any],
    *,
    is_leaf: Callable[[Any], None] | None = None,
):

    if not isinstance(cond, Callable):
        # a callable that accepts a leaf and returns a boolean
        # but *not* a tree with the same structure as tree with boolean values.
        raise TypeError(
            f"`cond` must be a callable that accepts a leaf and returns a boolean "
            f" Got {cond=} and {tree=}."
        )

    treelib = sepes._src.backend.treelib

    def map_func(x):
        return func(x) if cond(x) else x

    return treelib.map(map_func, tree, is_leaf=is_leaf)


def tree_mask(
    tree: T,
    cond: Callable[[Any], bool] = is_nondiff,
    *,
    is_leaf: Callable[[Any], None] | None = None,
):
    """Mask leaves of a pytree based on ``mask`` boolean pytree or callable.

    Masked leaves are wrapped with a wrapper that yields no leaves when
    ``tree_flatten`` is called on it.

    Args:
        tree: A pytree of values.
        cond: A callable that accepts a leaf and returns a boolean to mark the leaf
            for masking. Defaults to masking non-differentiable leaf nodes that
            are not instances of of python float, python complex, or inexact
            array types.
        is_leaf: A callable that accepts a leaf and returns a boolean. If
            provided, it is used to determine if a value is a leaf. for example,
            ``is_leaf=lambda x: isinstance(x, list)`` will treat lists as leaves
            and will not recurse into them.

    Example:
        >>> import sepes as sp
        >>> import jax
        >>> tree = [1, 2, {"a": 3, "b": 4.}]
        >>> # mask all non-differentiable nodes by default
        >>> masked_tree = sp.tree_mask(tree)
        >>> masked_tree
        [#1, #2, {'a': #3, 'b': 4.0}]
        >>> jax.tree_util.tree_leaves(masked_tree)
        [4.0]
        >>> sp.tree_unmask(masked_tree)
        [1, 2, {'a': 3, 'b': 4.0}]

    Example:
        Pass non-differentiable values to ``jax.grad``

        >>> import sepes as sp
        >>> import jax
        >>> @jax.grad
        ... def square(tree):
        ...     tree = sp.tree_unmask(tree)
        ...     return tree[0] ** 2
        >>> tree = (1., 2)  # contains a non-differentiable node
        >>> square(sp.tree_mask(tree))
        (Array(2., dtype=float32, weak_type=True), #2)
    """
    return _tree_mask_map(tree, cond=cond, func=mask, is_leaf=is_leaf)


def tree_unmask(tree: T, cond: Callable[[Any], bool] = lambda _: True):
    """Undo the masking of tree leaves according to ``cond``. defaults to unmasking all leaves.

    Args:
        tree: A pytree of values.
        cond: A callable that accepts a leaf and returns a boolean to mark the
            leaf to be unmasked. Defaults to always unmask.

    Example:
        >>> import sepes as sp
        >>> import jax
        >>> tree = [1, 2, {"a": 3, "b": 4.}]
        >>> # mask all non-differentiable nodes by default
        >>> masked_tree = sp.tree_mask(tree)
        >>> masked_tree
        [#1, #2, {'a': #3, 'b': 4.0}]
        >>> jax.tree_util.tree_leaves(masked_tree)
        [4.0]
        >>> sp.tree_unmask(masked_tree)
        [1, 2, {'a': 3, 'b': 4.0}]

    Example:
        Pass non-differentiable values to ``jax.grad``

        >>> import sepes as sp
        >>> import jax
        >>> @jax.grad
        ... def square(tree):
        ...     tree = sp.tree_unmask(tree)
        ...     return tree[0] ** 2
        >>> tree = (1., 2)  # contains a non-differentiable node
        >>> square(sp.tree_mask(tree))
        (Array(2., dtype=float32, weak_type=True), #2)
    """
    return _tree_mask_map(tree, cond=cond, func=unmask, is_leaf=is_masked)


if is_package_avaiable("jax"):
    import jax

    # do not touch jax.core.Tracer instances.
    # otherwise calling `freeze` inside a jax transformation on
    # a tracer will hide the tracer from jax and will cause leaked tracer
    # error.
    @mask.def_type(jax.core.Tracer)
    def _(value: jax.core.Tracer) -> jax.core.Tracer:
        return value



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
from collections import namedtuple
from typing import NamedTuple

import pytest

from sepes._src.backend import arraylib, backend, treelib
from sepes._src.code_build import autoinit
from sepes._src.tree_base import TreeClass, _mutable_instance_registry
from sepes._src.tree_index import at, BaseKey
from sepes._src.tree_util import is_tree_equal, leafwise, value_and_tree

test_arraylib = os.environ.get("SEPES_TEST_ARRAYLIB", "numpy")

if test_arraylib == "jax":
    import jax.numpy as arraylib

    default_int = arraylib.int32
elif test_arraylib in ["numpy", "default"]:
    import numpy as arraylib

    default_int = arraylib.int64
elif test_arraylib == "torch":
    import torch as arraylib

    arraylib.array = arraylib.tensor
    default_int = arraylib.int64
else:
    raise ImportError("no backend installed")


@leafwise
class ClassTree(TreeClass):
    """Tree class for testing."""

    def __init__(self, a: int, b: dict, e: int):
        """Initialize."""
        self.a = a
        self.b = b
        self.e = e


@leafwise
class ClassSubTree(TreeClass):
    """Tree class for testing."""

    def __init__(self, c: int, d: int):
        """Initialize."""
        self.c = c
        self.d = d


# key
tree1 = dict(a=1, b=dict(c=2, d=3), e=4)
tree2 = ClassTree(1, dict(c=2, d=3), 4)
tree3 = ClassTree(1, ClassSubTree(2, 3), 4)

# index
tree4 = [1, [2, 3], 4]
tree5 = (1, (2, 3), 4)
tree6 = [1, ClassSubTree(2, 3), 4]

# mixed
tree7 = dict(a=1, b=[2, 3], c=4)
tree8 = dict(a=1, b=ClassSubTree(c=2, d=3), e=4)

# by mask
tree9 = ClassTree(1, dict(c=2, d=3), arraylib.array([4, 5, 6]))

_X = 1_000


@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, dict(a=None, b=dict(c=2, d=None), e=None), ("b", "c")],
        [tree2, ClassTree(None, dict(c=2, d=None), None), ("b", "c")],
        [tree3, ClassTree(None, ClassSubTree(2, None), None), ("b", "c")],
        # by index
        [tree3, ClassTree(None, ClassSubTree(c=2, d=None), None), (1, 0)],
        [tree4, [None, [2, None], None], (1, 0)],
        [tree5, (None, (2, None), None), (1, 0)],
        [tree6, [None, ClassSubTree(2, None), None], (1, 0)],
        # mixed
        [tree7, dict(a=None, b=[2, None], c=None), ("b", 0)],
        [tree8, dict(a=None, b=ClassSubTree(c=2, d=None), e=None), ("b", 0)],
        # by regex
        [tree1, dict(a=None, b=dict(c=2, d=None), e=None), ("b", re.compile("c"))],
        [tree2, ClassTree(None, dict(c=2, d=None), None), ("b", re.compile("c"))],
        [tree3, ClassTree(None, ClassSubTree(2, None), None), ("b", re.compile("c"))],
        # by ellipsis
        [tree1, tree1, (...,)],
        [tree2, tree2, (...,)],
        [tree3, tree3, (...,)],
        [tree4, tree4, (...,)],
        [tree5, tree5, (...,)],
        [tree6, tree6, (...,)],
        [tree7, tree7, (...,)],
        [tree8, tree8, (...,)],
    ],
)
def test_indexer_get(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.get(), expected)
    assert is_tree_equal(indexer.get(is_parallel=True), expected)


@pytest.mark.skipif(backend == "default", reason="no array backend installed")
@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by boolean mask
        [
            tree9,
            ClassTree(None, dict(c=None, d=3), arraylib.array([4, 5, 6])),
            (tree9 > 2,),
        ],
        [
            tree9,
            ClassTree(None, dict(c=None, d=None), arraylib.array([5, 6])),
            (tree9 > 4,),
        ],
        [tree9, tree9, (tree9 == tree9,)],
        [
            tree9,
            ClassTree(
                None, dict(c=None, d=None), arraylib.array([], dtype=default_int)
            ),
            (tree9 != tree9,),
        ],
        # by ellipsis
        [tree9, tree9, (...,)],
    ],
)
def test_array_indexer_get(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.get(), expected)
    assert is_tree_equal(indexer.get(is_parallel=True), expected)


@pytest.mark.skipif(backend != "jax", reason="test jax jit with get")
def test_get_fill_value():
    import jax
    import jax.numpy as jnp

    tree = dict(a=jnp.array([1, 2, 3]), b=jnp.array([4, 5, 6]))
    mask = dict(
        a=jnp.array([False, True, False]),
        b=jnp.array([False, True, False]),
    )

    @jax.jit
    def jit_func(tree):
        return at(tree)[mask].get(fill_value=0)

    out = jit_func(tree)
    a = out["a"]
    b = out["b"]
    assert jnp.all(a == jnp.array([0, 2, 0]))
    assert jnp.all(b == jnp.array([0, 5, 0]))


@pytest.mark.parametrize(
    ["tree", "expected", "where", "set_value"],
    [
        # by name
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", "c"), _X],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", "c"), _X],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", "c"), _X],
        # by index
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), (1, 0), _X],
        [tree4, [1, [_X, 3], 4], (1, 0), _X],
        [tree5, (1, (_X, 3), 4), (1, 0), _X],
        [tree6, [1, ClassSubTree(_X, 3), 4], (1, 0), _X],
        # mixed
        [tree7, dict(a=1, b=[2, _X], c=4), ("b", 1), _X],
        [tree8, dict(a=1, b=ClassSubTree(c=2, d=_X), e=4), ("b", 1), _X],
        # by regex
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", re.compile("c")), _X],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", re.compile("c")), _X],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", re.compile("c")), _X],
        # by ellipsis
        [
            tree1,
            dict(a=_X, b=dict(c=_X, d=_X), e=_X),
            (...,),
            dict(a=_X, b=dict(c=_X, d=_X), e=_X),
        ],
        [tree2, ClassTree(_X, dict(c=_X, d=_X), _X), (...,), _X],
        [tree3, ClassTree(_X, ClassSubTree(_X, _X), _X), (...,), _X],
        [tree4, [_X, [_X, _X], _X], (...,), _X],
        [tree5, (_X, (_X, _X), _X), (...,), _X],
        [tree6, [_X, ClassSubTree(_X, _X), _X], (...,), _X],
        [tree7, dict(a=_X, b=[_X, _X], c=_X), (...,), _X],
        [tree8, dict(a=_X, b=ClassSubTree(c=_X, d=_X), e=_X), (...,), _X],
    ],
)
def test_indexer_set(tree, expected, where, set_value):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.set(set_value), expected)
    assert is_tree_equal(indexer.set(set_value, is_parallel=True), expected)


@pytest.mark.skipif(backend == "default", reason="no array backend installed")
@pytest.mark.parametrize(
    ["tree", "expected", "where", "set_value"],
    [
        # by name
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", "c"), _X],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", "c"), _X],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", "c"), _X],
        # by index
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), (1, 0), _X],
        [tree4, [1, [_X, 3], 4], (1, 0), _X],
        [tree5, (1, (_X, 3), 4), (1, 0), _X],
        [tree6, [1, ClassSubTree(_X, 3), 4], (1, 0), _X],
        # mixed
        [tree7, dict(a=1, b=[2, _X], c=4), ("b", 1), _X],
        [tree8, dict(a=1, b=ClassSubTree(c=2, d=_X), e=4), ("b", 1), _X],
        # by regex
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", re.compile("c")), _X],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", re.compile("c")), _X],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", re.compile("c")), _X],
        # by ellipsis
        [
            tree1,
            dict(a=_X, b=dict(c=_X, d=_X), e=_X),
            (...,),
            dict(a=_X, b=dict(c=_X, d=_X), e=_X),
        ],
        [tree2, ClassTree(_X, dict(c=_X, d=_X), _X), (...,), _X],
        [tree3, ClassTree(_X, ClassSubTree(_X, _X), _X), (...,), _X],
        [tree4, [_X, [_X, _X], _X], (...,), _X],
        [tree5, (_X, (_X, _X), _X), (...,), _X],
        [tree6, [_X, ClassSubTree(_X, _X), _X], (...,), _X],
        [tree7, dict(a=_X, b=[_X, _X], c=_X), (...,), _X],
        [tree8, dict(a=_X, b=ClassSubTree(c=_X, d=_X), e=_X), (...,), _X],
    ],
)
def test_array_indexer_set(tree, expected, where, set_value):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.set(set_value), expected)
    assert is_tree_equal(indexer.set(set_value, is_parallel=True), expected)


@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", "c")],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", "c")],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", "c")],
        # by index
        [tree4, [1, [_X, 3], 4], (1, 0)],
        [tree5, (1, (_X, 3), 4), (1, 0)],
        [tree6, [1, ClassSubTree(_X, 3), 4], (1, 0)],
        # mixed
        [tree7, dict(a=1, b=[2, _X], c=4), ("b", 1)],
        [tree8, dict(a=1, b=ClassSubTree(c=2, d=_X), e=4), ("b", 1)],
        # by regex
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", re.compile("c"))],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", re.compile("c"))],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", re.compile("c"))],
        # by ellipsis
        [tree1, dict(a=_X, b=dict(c=_X, d=_X), e=_X), (...,)],
        [tree2, ClassTree(_X, dict(c=_X, d=_X), _X), (...,)],
        [tree3, ClassTree(_X, ClassSubTree(_X, _X), _X), (...,)],
        [tree4, [_X, [_X, _X], _X], (...,)],
        [tree5, (_X, (_X, _X), _X), (...,)],
        [tree6, [_X, ClassSubTree(_X, _X), _X], (...,)],
        [tree7, dict(a=_X, b=[_X, _X], c=_X), (...,)],
        [tree8, dict(a=_X, b=ClassSubTree(c=_X, d=_X), e=_X), (...,)],
    ],
)
def test_indexer_apply(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.apply(lambda _: _X), expected)
    assert is_tree_equal(
        indexer.apply(lambda _: _X, is_parallel=True),
        expected,
    )


@pytest.mark.skipif(backend == "default", reason="no array backend installed")
@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", "c")],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", "c")],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", "c")],
        # by index
        [tree4, [1, [_X, 3], 4], (1, 0)],
        [tree5, (1, (_X, 3), 4), (1, 0)],
        [tree6, [1, ClassSubTree(_X, 3), 4], (1, 0)],
        # mixed
        [tree7, dict(a=1, b=[2, _X], c=4), ("b", 1)],
        [tree8, dict(a=1, b=ClassSubTree(c=2, d=_X), e=4), ("b", 1)],
        # by regex
        [tree1, dict(a=1, b=dict(c=_X, d=3), e=4), ("b", re.compile("c"))],
        [tree2, ClassTree(1, dict(c=_X, d=3), 4), ("b", re.compile("c"))],
        [tree3, ClassTree(1, ClassSubTree(_X, 3), 4), ("b", re.compile("c"))],
        # by ellipsis
        [tree1, dict(a=_X, b=dict(c=_X, d=_X), e=_X), (...,)],
        [tree2, ClassTree(_X, dict(c=_X, d=_X), _X), (...,)],
        [tree3, ClassTree(_X, ClassSubTree(_X, _X), _X), (...,)],
        [tree4, [_X, [_X, _X], _X], (...,)],
        [tree5, (_X, (_X, _X), _X), (...,)],
        [tree6, [_X, ClassSubTree(_X, _X), _X], (...,)],
        [tree7, dict(a=_X, b=[_X, _X], c=_X), (...,)],
        [tree8, dict(a=_X, b=ClassSubTree(c=_X, d=_X), e=_X), (...,)],
    ],
)
def test_array_indexer_apply(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(indexer.apply(lambda _: _X), expected)
    assert is_tree_equal(
        indexer.apply(lambda _: _X, is_parallel=True),
        expected,
    )


@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, 5, ("b", ("c", "d"))],
        [tree2, 5, ("b", ("c", "d"))],
        [tree3, 5, ("b", ("c", "d"))],
        # by index
        [tree4, 5, (1, (0, 1))],
        [tree5, 5, (1, (0, 1))],
        # mixed
        [tree7, 5, ("b", (0, 1))],
        # by regex
        [tree1, 5, ("b", re.compile("c|d"))],
        [tree2, 5, ("b", re.compile("c|d"))],
        [tree3, 5, ("b", re.compile("c|d"))],
        # by ellipsis
        [tree1, 1 + 2 + 3 + 4, (...,)],
        [tree2, 1 + 2 + 3 + 4, (...,)],
        [tree3, 1 + 2 + 3 + 4, (...,)],
        [tree4, 1 + 2 + 3 + 4, (...,)],
        [tree5, 1 + 2 + 3 + 4, (...,)],
        [tree6, 1 + 2 + 3 + 4, (...,)],
        [tree7, 1 + 2 + 3 + 4, (...,)],
        [tree8, 1 + 2 + 3 + 4, (...,)],
    ],
)
def test_indexer_reduce(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(
        indexer.reduce(lambda x, y: x + y, initializer=0),
        expected,
    )


@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, 5, ("b", ("c", "d"))],
        [tree2, 5, ("b", ("c", "d"))],
        [tree3, 5, ("b", ("c", "d"))],
        # by index
        [tree4, 5, (1, (0, 1))],
        [tree5, 5, (1, (0, 1))],
        # mixed
        [tree7, 5, ("b", (0, 1))],
        # by regex
        [tree1, 5, ("b", re.compile("c|d"))],
        [tree2, 5, ("b", re.compile("c|d"))],
        [tree3, 5, ("b", re.compile("c|d"))],
        # by ellipsis
        [tree1, 1 + 2 + 3 + 4, (...,)],
        [tree2, 1 + 2 + 3 + 4, (...,)],
        [tree3, 1 + 2 + 3 + 4, (...,)],
        [tree4, 1 + 2 + 3 + 4, (...,)],
        [tree5, 1 + 2 + 3 + 4, (...,)],
        [tree6, 1 + 2 + 3 + 4, (...,)],
        [tree7, 1 + 2 + 3 + 4, (...,)],
        [tree8, 1 + 2 + 3 + 4, (...,)],
    ],
)
def test_array_indexer_reduce(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(
        indexer.reduce(lambda x, y: x + y, initializer=0),
        expected,
    )


@pytest.mark.parametrize(
    ["tree", "expected", "where"],
    [
        # by name
        [tree1, (dict(a=1, b=dict(c=2, d=5), e=4), 3), ("b", ("c", "d"))],
        [tree2, (ClassTree(1, dict(c=2, d=5), 4), 3), ("b", ("c", "d"))],
        [tree3, (ClassTree(1, ClassSubTree(2, 5), 4), 3), ("b", ("c", "d"))],
        # by index
        [tree4, ([1, [2, 5], 4], 3), (1, (0, 1))],
        [tree5, ((1, (2, 5), 4), 3), (1, (0, 1))],
        # mixed
        [tree7, (dict(a=1, b=[2, 5], c=4), 3), ("b", (0, 1))],
        # [tree8, (dict(a=1, b=ClassSubTree(c=2, d=5), e=4), 3), ("b", (0, 1))],
        # by regex
        [tree1, (dict(a=1, b=dict(c=2, d=5), e=4), 3), ("b", re.compile("c|d"))],
        [tree2, (ClassTree(1, dict(c=2, d=5), 4), 3), ("b", re.compile("c|d"))],
        [tree3, (ClassTree(1, ClassSubTree(2, 5), 4), 3), ("b", re.compile("c|d"))],
    ],
)
def test_indexer_scan(tree, expected, where):
    indexer = at(tree, where=where)
    assert is_tree_equal(
        indexer.scan(lambda x, s: (x + s, x), state=0),
        expected,
    )


def test_method_call():
    @leafwise
    @autoinit
    class Tree(TreeClass):
        a: int = 1

        def increment(self):
            self.a += 1

        def show(self):
            return 1

    t = Tree()

    @autoinit
    class Tree2(TreeClass):
        b: Tree = Tree()

    assert is_tree_equal(value_and_tree(lambda T: T.increment())(t)[1], Tree(2))
    assert is_tree_equal(value_and_tree(lambda T: T.b.show())(Tree2())[0], 1)

    with pytest.raises(AttributeError):
        value_and_tree(t.bla)()

    with pytest.raises(TypeError):
        value_and_tree(t.a)()

    @leafwise
    @autoinit
    class A(TreeClass):
        a: int

        def __call__(self, x):
            self.a += x
            return x

    a = A(1)
    _, b = value_and_tree(lambda A: A(2))(a)

    assert treelib.flatten(a)[0] == [1]
    assert treelib.flatten(b)[0] == [3]

    with pytest.raises(TypeError):
        a.at[0](1)


def test_call_context():
    @autoinit
    class L2(TreeClass):
        a: int = 1

        def delete(self, name):
            del self.a

    t = L2()

    _mutable_instance_registry.add(id(t))
    t.delete("a")
    _mutable_instance_registry.discard(id(t))

    with pytest.raises(AttributeError):
        t.delete("a")


@pytest.mark.parametrize("where", [("a", [1]), (0, [1])])
def test_unsupported_where(where):
    t = namedtuple("a", ["x", "y"])(1, 2)
    with pytest.raises(NotImplementedError):
        at(t, where=where).get()


@pytest.mark.skipif(backend != "jax", reason="jax backend needed")
def test_custom_key_jax():
    class NameTypeContainer(NamedTuple):
        name: str
        type: type

    class Tree:
        def __init__(self, a, b) -> None:
            self.a = a
            self.b = b

        @property
        def at(self):
            return at(self)

    if backend == "jax":
        import jax.tree_util as jtu

        def tree_flatten(tree):
            return (tree.a, tree.b), None

        def tree_unflatten(aux_data, children):
            return Tree(*children)

        def tree_flatten_with_keys(tree):
            ak = (NameTypeContainer("a", type(tree.a)), tree.a)
            bk = (NameTypeContainer("b", type(tree.b)), tree.b)
            return (ak, bk), None

        jtu.register_pytree_with_keys(
            nodetype=Tree,
            flatten_func=tree_flatten,
            flatten_with_keys=tree_flatten_with_keys,
            unflatten_func=tree_unflatten,
        )


@pytest.mark.skipif(backend not in ["torch", "numpy"], reason="optree backend needed")
def test_custom_key_optreee():
    class NameTypeContainer(NamedTuple):
        name: str
        type: type

    class Tree:
        def __init__(self, a, b) -> None:
            self.a = a
            self.b = b

        @property
        def at(self):
            return at(self)

    import optree as ot

    def tree_flatten(tree):
        ka = NameTypeContainer("a", type(tree.a))
        kb = NameTypeContainer("b", type(tree.b))
        return (tree.a, tree.b), None, (ka, kb)

    def tree_unflatten(aux_data, children):
        return Tree(*children)

    ot.register_pytree_node(
        Tree,
        flatten_func=tree_flatten,
        unflatten_func=tree_unflatten,
        namespace="sepes",
    )

    tree = Tree(1, 2)

    class MatchNameType(BaseKey):
        def __init__(self, name, type):
            self.name = name
            self.type = type

        def __eq__(self, other):
            if isinstance(other, NameTypeContainer):
                return other == (self.name, self.type)
            return False

    assert treelib.tree_flatten(tree.at[MatchNameType("a", int)].get())[0] == [1]


def test_repr_str():
    @autoinit
    class Tree(TreeClass):
        a: int = 1
        b: int = 2

    t = Tree()

    assert repr(t.at["a"]) == "at(Tree(a=1, b=2), where=['a'])"
    assert str(t.at["a"]) == "at(Tree(a=1, b=2), where=['a'])"
    assert repr(t.at[...]) == "at(Tree(a=1, b=2), where=[Ellipsis])"


def test_compat_mask():
    tree = [1, 2, [3, 4]]
    tree_ = at(tree)[[False, False, True]].set(10)
    assert tree_ == [1, 2, 10]


def test_pluck():
    tree = [1, 2, [3, 4]]
    subtrees = at(tree)[2].pluck()
    assert subtrees[0] == [3, 4]
    assert at(tree)[0, 1].pluck(1) == [1]
    assert at(tree)[0, 1].pluck(2) == [1, 2]

    tree = dict(a=1, b=2)
    assert at(tree)[...].pluck() == [1, 2]


@pytest.mark.skipif(backend != "jax", reason="jax backend needed")
def test_call():
    import jax.tree_util as jtu

    @jtu.register_pytree_with_keys_class
    class Counter:
        def __init__(self, count: int):
            self.count = count

        def tree_flatten_with_keys(self):
            return (["count", self.count],), None

        @classmethod
        def tree_unflatten(cls, aux_data, children):
            del aux_data
            return cls(*children)

        def increment_count(self) -> int:
            # mutates the tree
            self.count += 1
            return self.count

        def __repr__(self) -> str:
            return f"Tree(count={self.count})"

    counter = Counter(0)
    cur_count, new_counter = value_and_tree(lambda C: C.increment_count())(counter)
    assert counter.count == 0
    assert cur_count == 1
    assert new_counter.count == 1
    assert not (counter is new_counter)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for pretty printing pytrees."""

from __future__ import annotations

import functools as ft
import inspect
import math
from itertools import zip_longest
from math import prod
from types import FunctionType
from typing import Any, Callable, NamedTuple, Sequence

from typing_extensions import TypeAlias, TypedDict, Unpack

import sepes
import sepes._src.backend.arraylib as arraylib
from sepes._src.backend import is_package_avaiable
from sepes._src.tree_util import (
    Node,
    construct_tree,
    is_path_leaf_depth_factory,
    tree_type_path_leaves,
)


class PPSpec(TypedDict):
    indent: int
    width: int
    depth: int | float


PyTree = Any

PP = Callable[[Any, Unpack[PPSpec]], str]


class ShapeDTypePP(NamedTuple):
    shape: tuple[int, ...]
    dtype: Any


def pp(printer, node: Any, **spec: Unpack[PPSpec]) -> str:
    return (
        "..."
        if spec["depth"] < 0
        else format_width(printer.dispatch(node, **spec), width=spec["width"])
    )


def pps(pp: PP, xs: Sequence[Any], **spec: Unpack[PPSpec]) -> str:
    if spec["depth"] < 1:
        return "..."

    spec["indent"] += 1
    spec["depth"] -= 1

    text = (
        "\n"
        + "\t" * spec["indent"]
        + (", \n" + "\t" * spec["indent"]).join(pp(x, **spec) for x in xs)
        + "\n"
        + "\t" * (spec["indent"] - 1)
    )

    return format_width(text, width=spec["width"])


def kv_pp(pp: PP, x: tuple[str, Any], **spec: Unpack[PPSpec]) -> str:
    return f"{x[0]}:{pp(x[1], **spec)}"


def av_pp(pp: PP, x: tuple[str, Any], **spec: Unpack[PPSpec]) -> str:
    return f"{x[0]}={pp(x[1], **spec)}"


def tree_repr(
    tree: PyTree,
    *,
    width: int = 80,
    tabwidth: int = 2,
    depth: int | float = float("inf"),
):
    out = tree_repr.dispatch(tree, indent=0, width=width, depth=depth)
    return out.expandtabs(tabwidth)


def tree_str(
    tree: PyTree,
    *,
    width: int = 80,
    tabwidth: int = 2,
    depth: int | float = float("inf"),
):
    out = tree_str.dispatch(tree, indent=0, width=width, depth=depth)
    return out.expandtabs(tabwidth)


def repr_dispatch(node: Any, **spec: Unpack[PPSpec]) -> str:
    return (
        ("\n" + "\t" * (spec["indent"])).join(text.split("\n"))
        if "\n" in (text := repr(node))
        else text
    )


def str_dispatch(node: Any, **spec: Unpack[PPSpec]) -> str:
    return (
        ("\n" + "\t" * (spec["indent"])).join(text.split("\n"))
        if "\n" in (text := str(node))
        else text
    )


tree_repr.dispatch = ft.singledispatch(repr_dispatch)
tree_repr.def_type = tree_repr.dispatch.register
tree_repr.pps = pps
tree_repr.pp = ft.partial(pp, tree_repr)
tree_repr.kv_pp = ft.partial(kv_pp, tree_repr.pp)
tree_repr.av_pp = ft.partial(av_pp, tree_repr.pp)

tree_str.dispatch = ft.singledispatch(str_dispatch)
tree_str.def_type = tree_str.dispatch.register
tree_str.pps = pps
tree_str.pp = ft.partial(pp, tree_str)
tree_str.kv_pp = ft.partial(kv_pp, tree_str.pp)
tree_str.av_pp = ft.partial(av_pp, tree_str.pp)


@tree_repr.def_type(ShapeDTypePP)
@tree_str.def_type(ShapeDTypePP)
def _(node: Any, **spec: Unpack[PPSpec]) -> str:
    """Pretty print a node with dtype and shape."""
    shape = f"{node.shape}".replace(",", "")
    shape = shape.replace("(", "[")
    shape = shape.replace(")", "]")
    shape = shape.replace(" ", ",")
    dtype = f"{node.dtype}".replace("int", "i")
    dtype = dtype.replace("float", "f")
    dtype = dtype.replace("complex", "c")
    return dtype + shape


@tree_repr.def_type(FunctionType)
@tree_str.def_type(FunctionType)
def _(func: Callable, **spec: Unpack[PPSpec]) -> str:
    del spec
    fullargspec = inspect.getfullargspec(func)

    header: list[str] = []

    if len(fullargspec.args):
        header += fullargspec.args
    if fullargspec.varargs is not None:
        header += ["*" + fullargspec.varargs]
    if len(fullargspec.kwonlyargs):
        if fullargspec.varargs is None:
            header += ["*"]
        header += fullargspec.kwonlyargs
    if fullargspec.varkw is not None:
        header += ["**" + fullargspec.varkw]

    *_, name = getattr(func, "__qualname__", "").split(".")
    return f"{name}({', '.join(header)})"


@tree_str.def_type(ft.partial)
def _(node: ft.partial, **spec: Unpack[PPSpec]) -> str:
    func = tree_str.pp(node.func, **spec)
    args = tree_str.pps(tree_str.pp, node.args, **spec)
    keywords = tree_str.pps(tree_str.kv_pp, node.keywords, **spec)
    return "partial(" + ",".join([func, args, keywords]) + ")"


@tree_str.def_type(list)
def _(node: list, **spec: Unpack[PPSpec]) -> str:
    return "[" + tree_str.pps(tree_str.pp, node, **spec) + "]"


@tree_str.def_type(tuple)
def _(node: tuple, **spec: Unpack[PPSpec]) -> str:
    if not hasattr(node, "_fields"):
        return "(" + tree_str.pps(tree_str.pp, node, **spec) + ")"
    name = type(node).__name__
    kvs = node._asdict().items()
    return name + "(" + tree_str.pps(tree_str.av_pp, kvs, **spec) + ")"


@tree_str.def_type(set)
def _(node: set, **spec: Unpack[PPSpec]) -> str:
    return "{" + tree_str.pps(tree_str.pp, node, **spec) + "}"


@tree_str.def_type(dict)
def _(node: dict, **spec: Unpack[PPSpec]) -> str:
    name = type(node).__name__
    return name + "(" + tree_str.pps(tree_str.av_pp, node.items(), **spec) + ")"


@tree_repr.def_type(str)
@tree_str.def_type(str)
def _(node: str, **spec: Unpack[PPSpec]) -> str:
    return node


for ndarray in arraylib.ndarrays:

    @tree_repr.def_type(ndarray)
    def array_pp(node, **spec: Unpack[PPSpec]) -> str:
        shape_dtype = ShapeDTypePP(arraylib.shape(node), arraylib.dtype(node))
        base = tree_repr.pp(shape_dtype, **spec)

        if not (arraylib.is_floating(node) or arraylib.is_integer(node)):
            return base
        if prod(arraylib.shape(node)) == 0:
            return base

        # Extended repr for numpy array, with extended information
        # this part of the function is inspired by
        # lovely-jax https://github.com/xl0/lovely-jax
        L, H = arraylib.min(node), arraylib.max(node)
        interval = "(" if math.isinf(L) else "["
        interval += f"{L},{H}" if arraylib.is_integer(node) else f"{L:.2f},{H:.2f}"
        interval += ")" if math.isinf(H) else "]"
        interval = interval.replace("inf", "∞")
        mean, std = f"{arraylib.mean(node):.2f}", f"{arraylib.std(node):.2f}"
        return f"{base}(μ={mean}, σ={std}, ∈{interval})"


@tree_repr.def_type(ft.partial)
def _(node: ft.partial, **spec: Unpack[PPSpec]) -> str:
    func = tree_repr.pp(node.func, **spec)
    args = tree_repr.pps(tree_repr.pp, node.args, **spec)
    keywords = tree_repr.pps(tree_repr.kv_pp, node.keywords, **spec)
    return "Partial(" + ",".join([func, args, keywords]) + ")"


@tree_repr.def_type(list)
def _(node: list, **spec: Unpack[PPSpec]) -> str:
    return "[" + tree_repr.pps(tree_repr.pp, node, **spec) + "]"


@tree_repr.def_type(tuple)
def _(node: tuple, **spec: Unpack[PPSpec]) -> str:
    if not hasattr(node, "_fields"):
        return "(" + tree_repr.pps(tree_repr.pp, node, **spec) + ")"
    name = type(node).__name__
    kvs = node._asdict().items()
    return name + "(" + tree_repr.pps(tree_repr.av_pp, kvs, **spec) + ")"


@tree_repr.def_type(set)
def _(node: set, **spec: Unpack[PPSpec]) -> str:
    return "{" + tree_repr.pps(tree_repr.pp, node, **spec) + "}"


@tree_repr.def_type(dict)
def _(node: dict, **spec: Unpack[PPSpec]) -> str:
    name = type(node).__name__
    return name + "(" + tree_repr.pps(tree_repr.av_pp, node.items(), **spec) + ")"


def tree_diagram(
    tree: Any,
    *,
    depth: int | float = float("inf"),
    is_leaf: Callable[[Any], None] | None = None,
    tabwidth: int = 4,
):
    """Pretty print arbitrary pytrees tree with tree structure diagram.

    Args:
        tree: arbitrary pytree.
        depth: depth of the tree to print. default is max depth.
        is_leaf: function to determine if a node is a leaf. default is None.
        tabwidth: tab width of the repr string. default is 4.

    Example:
        >>> import sepes as sp
        >>> @sp.autoinit
        ... class A(sp.TreeClass):
        ...     x: int = 10
        ...     y: int = (20,30)
        ...     z: int = 40

        >>> @sp.autoinit
        ... class B(sp.TreeClass):
        ...     a: int = 10
        ...     b: tuple = (20,30, A())

        >>> print(sp.tree_diagram(B(), depth=0))
        B(...)

        >>> print(sp.tree_diagram(B(), depth=1))
        B
        ├── .a=10
        └── .b=(...)


        >>> print(sp.tree_diagram(B(), depth=2))
        B
        ├── .a=10
        └── .b:tuple
            ├── [0]=20
            ├── [1]=30
            └── [2]=A(...)
    """
    vmark = ("│\t")[:tabwidth]  # vertical mark
    lmark = ("└" + "─" * (tabwidth - 2) + (" \t"))[:tabwidth]  # last mark
    cmark = ("├" + "─" * (tabwidth - 2) + (" \t"))[:tabwidth]  # connector mark
    smark = (" \t")[:tabwidth]  # space mark

    def step(
        node: Node,
        depth: int = 0,
        is_last: bool = False,
        is_lasts: tuple[bool, ...] = (),
    ) -> str:
        indent = "".join(smark if is_last else vmark for is_last in is_lasts[:-1])
        branch = (lmark if is_last else cmark) if depth > 0 else ""

        if (child_count := len(node.children)) == 0:
            (key, _), value = node.data
            text = f"{indent}"
            text += f"{branch}{key}=" if key is not None else ""
            text += tree_repr(value, depth=0)
            return text + "\n"

        (key, type), _ = node.data

        text = f"{indent}{branch}"
        text += f"{key}:" if key is not None else ""
        text += f"{type.__name__}\n"

        for i, child in enumerate(node.children.values()):
            text += step(
                child,
                depth=depth + 1,
                is_last=(i == child_count - 1),
                is_lasts=is_lasts + (i == child_count - 1,),
            )
        return text

    is_path_leaf = is_path_leaf_depth_factory(depth)
    root = construct_tree(tree, is_leaf=is_leaf, is_path_leaf=is_path_leaf)
    text = step(root, is_last=len(root.children) == 1)
    return (text if tabwidth is None else text.expandtabs(tabwidth)).rstrip()


def format_width(string, width=60):
    """Strip newline/tab characters if less than max width."""
    children_length = len(string) - string.count("\n") - string.count("\t")
    if children_length > width:
        return string
    return string.replace("\n", "").replace("\t", "")


# table printing

Row: TypeAlias = Sequence[str]  # list of columns


def _table(rows: list[Row]) -> str:
    """Generate a table from a list of rows."""

    def line(text: Row, widths: list[int]) -> str:
        return "\n".join(
            "│"
            + "│".join(col.ljust(width) for col, width in zip(line_row, widths))
            + "│"
            for line_row in zip_longest(*[t.split("\n") for t in text], fillvalue="")
        )

    widths = [max(map(len, "\n".join(col).split("\n"))) for col in zip(*rows)]
    spaces: Row = ["─" * width for width in widths]

    return (
        ("┌" + "┬".join(spaces) + "┐")
        + "\n"
        + ("\n├" + "┼".join(spaces) + "┤\n").join(line(row, widths) for row in rows)
        + "\n"
        + ("└" + "┴".join(spaces) + "┘")
    )


def size_pp(size: int, **spec: Unpack[PPSpec]):
    del spec
    order_alpha = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    size_order = int(math.log(size, 1024)) if size else 0
    text = f"{(size)/(1024**size_order):.2f}{order_alpha[size_order]}"
    return text


def tree_summary(
    tree: PyTree,
    *,
    depth: int | float = float("inf"),
    is_leaf: Callable[[Any], None] | None = None,
) -> str:
    """Print a summary of an arbitrary pytree.

    Args:
        tree: A pytree.
        depth: max depth to display the tree. Defaults to maximum depth.
        is_leaf: function to determine if a node is a leaf. Defaults to ``None``

    Returns:
        String summary of the tree structure:
            - First column: path to the node.
            - Second column: type of the node. to control the displayed type use
              ``tree_summary.def_type(type, func)`` to define a custom type display function.
            - Third column: number of leaves in the node. for arrays the number of leaves
              is the number of elements in the array, otherwise its 1. to control the
              number of leaves of a node use ``tree_summary.def_count(type,func)``
            - Fourth column: size of the node in bytes. if the node is array the size
              is the size of the array in bytes, otherwise the size is not displayed.
              to control the size of a node use ``tree_summary.def_size(type,func)``
            - Last row: type of parent, number of leaves of the parent

    Example:
        >>> import sepes as sp
        >>> import jax.numpy as jnp
        >>> print(sp.tree_summary([1, [2, [3]], jnp.array([1, 2, 3])]))
        ┌─────────┬────────────────────────────────────┬─────┬──────┐
        │Name     │Type                                │Count│Size  │
        ├─────────┼────────────────────────────────────┼─────┼──────┤
        │[0]      │int                                 │1    │      │
        ├─────────┼────────────────────────────────────┼─────┼──────┤
        │[1][0]   │int                                 │1    │      │
        ├─────────┼────────────────────────────────────┼─────┼──────┤
        │[1][1][0]│int                                 │1    │      │
        ├─────────┼────────────────────────────────────┼─────┼──────┤
        │[2]      │i32[3]                              │3    │12.00B│
        ├─────────┼────────────────────────────────────┼─────┼──────┤
        │Σ        │list[int,list[int,list[int]],i32[3]]│6    │12.00B│
        └─────────┴────────────────────────────────────┴─────┴──────┘

    Example:
        Display flops of a function in tree summary

        >>> import jax
        >>> import functools as ft
        >>> import sepes as sp
        >>> def count_flops(func, *args, **kwargs) -> int:
        ...     cost_analysis = jax.jit(func).lower(*args, **kwargs).cost_analysis()
        ...     return cost_analysis["flops"] if "flops" in cost_analysis else 0
        >>> class Flops:
        ...     def __init__(self, func, *args, **kwargs):
        ...         self.func = ft.partial(func, *args, **kwargs)
        >>> @sp.tree_summary.def_count(Flops)
        ... def _(node: Flops) -> int:
        ...     return count_flops(node.func)
        >>> @sp.tree_summary.def_type(Flops)
        ... def _(node: Flops) -> str:
        ...     return f"Flops({sp.tree_repr(node.func.func)})"
        >>> tree = dict(a=1, b=Flops(jax.nn.relu, jax.numpy.ones((10, 1))))
        >>> print(sp.tree_summary(tree))
        ┌─────┬───────────────────┬─────┬────┐
        │Name │Type               │Count│Size│
        ├─────┼───────────────────┼─────┼────┤
        │['a']│int                │1    │    │
        ├─────┼───────────────────┼─────┼────┤
        │['b']│Flops(jit(relu(x)))│10.0 │    │
        ├─────┼───────────────────┼─────┼────┤
        │Σ    │dict               │11.0 │    │
        └─────┴───────────────────┴─────┴────┘

    Example:
        Register custom type size rule

        >>> import jax
        >>> import sepes as sp
        >>> def func(x):
        ...     print(sp.tree_summary(x))
        ...     return x
        >>> class AbstractZero: ...
        >>> @sp.tree_summary.def_size(AbstractZero)
        ... def _(node: AbstractZero) -> int:
        ...     return 0
        >>> print(sp.tree_summary(AbstractZero()))
        ┌────┬────────────┬─────┬────┐
        │Name│Type        │Count│Size│
        ├────┼────────────┼─────┼────┤
        │Σ   │AbstractZero│1    │    │
        └────┴────────────┴─────┴────┘
    """
    treelib = sepes._src.backend.treelib
    empty_trace = ((), ())
    total_rows: list[list[str]] = [["Name", "Type", "Count", "Size"]]
    total_count = total_size = 0

    def tree_size(tree: PyTree) -> int:
        def reduce_func(acc, node):
            return acc + tree_summary.size_dispatcher(node)

        leaves, _ = treelib.flatten(tree)
        return ft.reduce(reduce_func, leaves, 0)

    def tree_count(tree: PyTree) -> int:
        def reduce_func(acc, node):
            return acc + tree_summary.count_dispatcher(node)

        leaves, _ = treelib.flatten(tree)
        return ft.reduce(reduce_func, leaves, 0)

    traces_leaves = tree_type_path_leaves(
        tree=tree,
        is_leaf=is_leaf,
        is_path_leaf=is_path_leaf_depth_factory(depth),
    )

    for trace, leaf in traces_leaves:
        total_count += (count := tree_count(leaf))
        total_size += (size := tree_size(leaf))

        if trace == empty_trace:
            continue

        paths, _ = trace
        path_string = treelib.keystr(paths)
        type_string = tree_summary.type_dispatcher(leaf)
        count_string = f"{count:,}" if count else ""
        size_string = size_pp(size) if size else ""
        total_rows += [[path_string, type_string, count_string, size_string]]

    path_string = "Σ"
    type_string = tree_summary.type_dispatcher(tree)
    count_string = f"{total_count:,}" if total_count else ""
    size_string = size_pp(total_size) if total_size else ""
    total_rows += [[path_string, type_string, count_string, size_string]]
    return _table(total_rows)


tree_summary.count_dispatcher = ft.singledispatch(lambda x: 1)
tree_summary.def_count = tree_summary.count_dispatcher.register
tree_summary.size_dispatcher = ft.singledispatch(lambda x: 0)
tree_summary.def_size = tree_summary.size_dispatcher.register
tree_summary.type_dispatcher = ft.singledispatch(lambda x: type(x).__name__)
tree_summary.def_type = tree_summary.type_dispatcher.register


for ndarray in arraylib.ndarrays:

    @tree_summary.def_size(ndarray)
    def _(node) -> int:
        return arraylib.nbytes(node)

    @tree_summary.def_count(ndarray)
    def _(node) -> int:
        return prod(arraylib.shape(node))

    @tree_summary.def_type(ndarray)
    def _(node: Any) -> str:
        """Return the type repr of the node."""
        shape = arraylib.shape(node)
        dtype = arraylib.dtype(node)
        return tree_repr(ShapeDTypePP(shape, dtype))

@tree_summary.def_type(list)
@tree_summary.def_type(tuple)
def _(node: tuple) -> str:
    # - output Container[types,...] instead of just container type in the type col.
    # - usually this encounterd if the tree_summary depth is not inf
    #   so the tree leaves could contain non-atomic types.
    treelib = sepes._src.backend.treelib

    one_level_types = treelib.map(
        tree_summary.type_dispatcher,
        node,
        is_leaf=lambda x: False if id(x) == id(node) else True,
    )
    return f"{type(node).__name__}[{','.join(one_level_types)}]"


if is_package_avaiable("jax"):
    # jax pretty printing extra handlers
    import jax

    @tree_str.def_type(jax.ShapeDtypeStruct)
    @tree_repr.def_type(jax.ShapeDtypeStruct)
    def _(node: jax.ShapeDtypeStruct, **spec: Unpack[PPSpec]) -> str:
        shape = arraylib.shape(node)
        dtype = arraylib.dtype(node)
        return tree_repr.dispatch(ShapeDTypePP(shape, dtype), **spec)

    # more readable repr for jax custom_jvp functions
    # for instance: jax.nn.relu is displayed as relu(x)
    # instead of <jax._src.custom_derivatives.custom_jvp object at ...>
    @tree_str.def_type(jax.custom_jvp)
    @tree_repr.def_type(jax.custom_jvp)
    def _(node: jax.custom_jvp, **spec: Unpack[PPSpec]) -> str:
        node = getattr(node, "__wrapped__", "<unknown>")
        return tree_repr.dispatch(node, **spec)

    @tree_str.def_type(type(jax.jit(lambda x: x)))
    @tree_repr.def_type(type(jax.jit(lambda x: x)))
    def _(node, **spec: Unpack[PPSpec]) -> str:
        # on copy pjit loses the __wrapped__ attribute (maybe a bug in jax)
        # so we need to handle it here
        node = getattr(node, "__wrapped__", "<unknown>")
        return "jit(" + tree_repr.dispatch(node, **spec) + ")"

    # without this rule, Tracer will be handled by the array handler
    # that display min/max/mean/std of the array.However `Tracer` does not
    # have these attributes so this will cause an error upon calculation.
    @tree_summary.def_type(jax.core.Tracer)
    @tree_repr.def_type(jax.core.Tracer)
    @tree_str.def_type(jax.core.Tracer)
    def _(node, **spec: Unpack[PPSpec]) -> str:
        shape = node.aval.shape
        dtype = node.aval.dtype
        string = tree_repr.dispatch(ShapeDTypePP(shape, dtype), **spec)
        return f"{type(node).__name__}({string})"

    # handle the sharding info if it is sharded
    @tree_summary.def_type(jax.Array)
    def _(node: Any) -> str:
        """Return the type repr of the node."""
        # global shape
        global_shape = arraylib.shape(node)
        shard_shape = node.sharding.shard_shape(global_shape)
        dtype = arraylib.dtype(node)
        global_info = tree_repr(ShapeDTypePP(global_shape, dtype))

        if global_shape == shard_shape:
            return global_info
        shard_info = tree_repr(ShapeDTypePP(shard_shape, dtype))
        return f"G:{global_info}\nS:{shard_info}"



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import Any, Callable, Hashable, Iterable, Literal, Tuple, TypedDict, TypeVar

# optree namespace identifier
namespace: str = os.environ.get("SEPES_NAMESPACE", "sepes")

Tree = TypeVar("Tree", bound=Any)
Leaf = TypeVar("Leaf", bound=Any)
KeyEntry = TypeVar("KeyEntry", bound=Hashable)
KeyPath = Tuple[KeyEntry, ...]
KeyPathLeaf = Tuple[KeyPath, Leaf]
pool_map = dict(thread=ThreadPoolExecutor, process=ProcessPoolExecutor)


class ParallelConfig(TypedDict):
    max_workers: int | None
    kind: Literal["thread", "process"]


def raise_future_execption(future):
    raise future.exception()


def concurrent_map(
    func: Callable[..., Any],
    flat: Iterable[Any],
    max_workers: int | None = None,
    kind: Literal["thread", "process"] = "thread",
) -> Iterable[Any]:
    with (executor := pool_map[kind](max_workers)) as executor:
        futures = [executor.submit(func, *args) for args in zip(*flat)]

    return [
        future.result()
        if future.exception() is None
        else raise_future_execption(future)
        for future in futures
    ]


class AbstractTreeLib(abc.ABC):
    """The minimal interface for tree operations used by sepes."""

    @staticmethod
    @abc.abstractmethod
    def map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], None] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def path_map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[Leaf], Any]:
        ...

    @staticmethod
    @abc.abstractmethod
    def path_flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[KeyPathLeaf], Any]:
        ...

    @staticmethod
    @abc.abstractmethod
    def unflatten(treedef: Any, leaves: Iterable[Any]) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def register_treeclass(klass: type[Tree]) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def register_static(klass: type[Tree]) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def attribute_key(name: str) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def sequence_key(index: int) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def dict_key(key: Hashable) -> Any:
        ...

    @staticmethod
    @abc.abstractmethod
    def keystr(keys: Any) -> str:
        ...



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from sepes._src.backend import backend
from sepes._src.code_build import autoinit
from sepes._src.tree_base import TreeClass
from sepes._src.tree_util import is_tree_equal, leafwise


@pytest.mark.skipif(backend != "jax", reason="jax backend is not installed")
def test_ops_with_jit():
    import jax
    import jax.numpy as arraylib

    @autoinit
    @leafwise
    class T0(TreeClass):
        a: jax.Array = arraylib.array(1)
        b: jax.Array = arraylib.array(2)
        c: jax.Array = arraylib.array(3)

    @autoinit
    @leafwise
    class T1(TreeClass):
        a: jax.Array = arraylib.array(1)
        b: jax.Array = arraylib.array(2)
        c: jax.Array = arraylib.array(3)
        d: jax.Array = arraylib.array([1, 2, 3])

    @jax.jit
    def getter(tree):
        return tree.at[...].get()

    @jax.jit
    def setter(tree):
        return tree.at[...].set(0)

    @jax.jit
    def applier(tree):
        return tree.at[...].apply(lambda _: 0)

    assert is_tree_equal(getter(T0()), T0())
    assert is_tree_equal(T0(0, 0, 0), setter(T0()))
    assert is_tree_equal(T0(0, 0, 0), applier(T0()))
    assert is_tree_equal(getter(T1()), T1())
    assert is_tree_equal(T1(0, 0, 0, 0), setter(T1()))
    assert is_tree_equal(T1(0, 0, 0, 0), applier(T1()))
    assert jax.jit(is_tree_equal)(T1(0, 0, 0, 0), applier(T1()))



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import functools as ft
import os
from importlib.util import find_spec
from typing import Literal, Callable
import logging
from contextlib import contextmanager


@ft.lru_cache(maxsize=None)
def is_package_avaiable(backend: str) -> bool:
    return find_spec(backend) is not None


# by importing the backend modules here, we register the backend implementations
# with the arraylib
if is_package_avaiable("torch"):
    import sepes._src.backend.arraylib.torch
if is_package_avaiable("jax"):
    import sepes._src.backend.arraylib.jax
if is_package_avaiable("numpy"):
    import sepes._src.backend.arraylib.numpy


def optree_backend():
    # no backend is available
    if not is_package_avaiable("optree"):
        raise ImportError("No backend is available. Please install `optree`.")
    from sepes._src.backend.treelib.optree import OpTreeTreeLib

    return OpTreeTreeLib()


def jax_backend():
    if not is_package_avaiable("jax"):
        raise ImportError("`jax` backend requires `jax` to be installed.")
    from sepes._src.backend.treelib.jax import JaxTreeLib

    return JaxTreeLib()


BackendLiteral = Literal["optree", "jax"]  # tree backend
backend: BackendLiteral = os.environ.get("SEPES_BACKEND", "default").lower()
backends_map: dict[BackendLiteral, Callable] = {}
backends_map["jax"] = jax_backend
backends_map["optree"] = optree_backend

if backend == "default":
    # backend promotion in essence is a search for the first available backend
    # in the following order: jax, optree
    # if no backend is available, then the default backend is used
    for backend_name in backends_map:
        if is_package_avaiable(backend_name):
            treelib = backends_map[backend_name]()
            backend = backend_name
            logging.info(f"Successfully set backend to `{backend_name}`")
            break
elif backend == "jax":
    treelib = jax_backend()
    logging.info("Successfully set backend to `jax`")
elif backend == "optree":
    treelib = optree_backend()
    logging.info(f"Successfully set backend to `{backend}`")
else:
    raise ValueError(f"Unknown backend: {backend!r}. available {backends_map.keys()=}")


@contextmanager
def backend_context(backend_name: BackendLiteral):
    """Context manager for switching the tree backend within a context.

    Args:
        backend_name: The name of the backend to switch to. available backends are
            ``optree`` and ``jax``.

    Example:
        Registering a custom tree class with optree backend:

        >>> import sepes as sp
        >>> import optree
        >>> with sp.backend_context("optree"):
        ...     class Tree(sp.TreeClass):
        ...         def __init__(self, a, b):
        ...             self.a = a
        ...             self.b = b
        ...     tree = Tree(1, 2)
        >>> optree.tree_flatten(tree, namespace="sepes")
        ([1, 2], PyTreeSpec(CustomTreeNode(Tree[('a', 'b')], [*, *]), namespace='sepes'))
    """
    global treelib, backend
    old_treelib = treelib
    old_backend = backend
    try:
        treelib = backends_map[backend_name]()
        yield
    finally:
        treelib = old_treelib
        backend = old_backend



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define a class that convert a class to a compatible tree structure."""

from __future__ import annotations

import abc
from typing import Any, Hashable, TypeVar

from typing_extensions import Self, Unpack

import sepes
from sepes._src.code_build import fields
from sepes._src.tree_index import at
from sepes._src.tree_pprint import PPSpec, tree_repr, tree_str
from sepes._src.tree_util import is_tree_equal, tree_copy, tree_hash, value_and_tree

T = TypeVar("T", bound=Hashable)
S = TypeVar("S")
PyTree = Any
EllipsisType = type(Ellipsis)  # TODO: use typing.EllipsisType when available
_mutable_instance_registry: set[int] = set()


def add_mutable_entry(node) -> None:
    _mutable_instance_registry.add(id(node))


def discard_mutable_entry(node) -> None:
    # use discard instead of remove to avoid raising KeyError
    # if the node has been removed in a parent node.
    _mutable_instance_registry.discard(id(node))


class TreeClassMeta(abc.ABCMeta):
    def __call__(klass: type[T], *a, **k) -> T:
        tree = getattr(klass, "__new__")(klass, *a, **k)
        # allow the setattr/delattr to set/delete attributes in the initialization
        # phase by flagging the instance as mutable.
        add_mutable_entry(tree)
        # initialize the instance with the instance marked as mutable.
        getattr(klass, "__init__")(tree, *a, **k)
        # remove the mutable flag after the initialization. to disallow
        # setattr/delattr to set/delete attributes after the initialization.
        discard_mutable_entry(tree)
        return tree


class TreeClass(metaclass=TreeClassMeta):
    """Convert a class to a pytree by inheriting from :class:`.TreeClass`.

    A pytree is any nested structure of containers and leaves. A container is
    a pytree can be a container or a leaf. Container examples are: a ``tuple``,
    ``list``, or ``dict``. A leaf is a non-container data structure like an
    ``int``, ``float``, ``string``, or ``Array``. :class:`.TreeClass` is a
    container pytree that holds other pytrees in its attributes.

    Note:
        :class:`.TreeClass` is immutable by default. This means that setting or
        deleting attributes after initialization is not allowed. This behavior
        is intended to prevent accidental mutation of the tree. All tree modifications
        on `TreeClass` are out-of-place. This means that all tree modifications
        return a new instance of the tree with the modified values.

        There are two ways to set or delete attributes after initialization:

        1. Using :attr:`.at` property to modify an *existing* leaf of the tree.

           >>> import sepes as sp
           >>> class Tree(sp.TreeClass):
           ...     def __init__(self, leaf: int):
           ...         self.leaf = leaf
           >>> tree = Tree(leaf=1)
           >>> new_tree = tree.at["leaf"].set(100)
           >>> tree is new_tree  # new instance is created
           False

        2. Using :func:`.value_and_tree` to call a method that mutates the tree.
           and apply the mutation on a *copy* of the tree. This option allows
           writing methods that mutate the tree instance but with these updates
           applied on a copy of the tree.

           >>> import sepes as sp
           >>> class Tree(sp.TreeClass):
           ...     def __init__(self, foo: int):
           ...         self.foo = foo
           ...     def add_bar(self, value:int) -> None:
           ...         # this method mutates the tree instance
           ...         # and will raise an `AttributeError` if called directly.
           ...         setattr(self, "bar", value)
           >>> tree = Tree(foo=1)
           >>> # now lets try to call `add_bar` directly
           >>> tree.add_bar(value=100)  # doctest: +SKIP
           Cannot set attribute value=100 to `key='bar'`  on an immutable instance of `Tree`.
           >>> output, tree_ = sp.value_and_tree(lambda T: T.add_bar(100))(tree)
           >>> tree, tree_
           (Tree(foo=1), Tree(foo=1, bar=100))

           This pattern is useful to write freely mutating methods, but with
           The expense of having to call through `at["method_name"]` instead of
           calling the method directly.

    Note:
        ``sepes`` offers two methods to construct the ``__init__`` method:

        1. Manual ``__init__`` method

           >>> import sepes as sp
           >>> class Tree(sp.TreeClass):
           ...     def __init__(self, a:int, b:float):
           ...         self.a = a
           ...         self.b = b
           >>> tree = Tree(a=1, b=2.0)

        2. Auto generated ``__init__`` method from type annotations.

           Using :func:`.autoinit` decorator where the type annotations are used to
           generate the ``__init__`` method. :func:`.autoinit`` with :func:`field`
           objects can be used to apply functions on the field values during
           initialization, support multiple argument kinds, and can apply functions
           on field values on getting the value. For more details see :func:`.autoinit`
           and :func:`.field`.

           >>> import sepes as sp
           >>> @sp.autoinit
           ... class Tree(sp.TreeClass):
           ...     a:int
           ...     b:float
           >>> tree = Tree(a=1, b=2.0)

    Note:
        Leaf-wise math operations are supported  using ``leafwise`` decorator.
        ``leafwise`` decorator adds ``__add__``, ``__sub__``, ``__mul__``, ... etc
        to registered pytrees. These methods apply math operations to each leaf of
        the tree. for example:

        >>> @sp.leafwise
        ... class Tree(sp.TreeClass):
        ...     def __init__(self, a:int, b:float):
        ...         self.a = a
        ...         self.b = b
        >>> tree = Tree(a=1, b=2.0)
        >>> tree + 1  # will add 1 to each leaf
        Tree(a=2, b=3.0)

    Note:
        Advanced indexing is supported using ``at`` property. Indexing can be
        used to ``get``, ``set``, or ``apply`` a function to a leaf or a group of
        leaves using ``leaf`` name, index or by a boolean mask.

        >>> class Tree(sp.TreeClass):
        ...     def __init__(self, a:int, b:float):
        ...         self.a = a
        ...         self.b = b
        >>> tree = Tree(a=1, b=2.0)
        >>> tree.at["a"].get()
        Tree(a=1, b=None)
        >>> tree.at[0].get()
        Tree(a=1, b=None)

    Note:
        ``AttributeError`` is raised, If a method that mutates the instance
        is called directly. Instead use :func:`.value_and_tree` to call
        the method on a copy of the tree. :func:`.value_and_tree` calls the function
        on copied input arguments to ensure non-mutating behavior.

        >>> import sepes as sp
        >>> class Counter(sp.TreeClass):
        ...     def __init__(self, count: int):
        ...         self.count = count
        ...     def increment(self, value):
        ...         self.count += value
        ...         return self.count
        >>> counter = Counter(0)
        >>> sp.value_and_tree(lambda C: C.increment(1))(counter)
        (1, Counter(count=1))

    Note:
        :class:`.TreeClass` inherits from ``abc.ABC`` meaning that it can `abc`
        features like ``@abc.abstractmethod`` can be used to define abstract
        behavior that can be implemented by subclasses.

    Warning:
        The structure should be organized as a tree. In essence, *cyclic references*
        are not allowed. The leaves of the tree are the values of the tree and
        the branches are the containers that hold the leaves.
    """

    def __init_subclass__(klass: type[T], **k):
        # disallow setattr/delattr to be overridden as they are used
        # to implement the immutable/controlled mutability behavior.
        if "__setattr__" in vars(klass):
            raise TypeError(f"Reserved method `__setattr__` defined in `{klass}`.")
        if "__delattr__" in vars(klass):
            raise TypeError(f"Reserved method `__delattr__` defined in `{klass}`.")
        super().__init_subclass__(**k)
        # - register the class with the proper tree backend.
        # - the registration envolves defining two rules: how to flatten the nested
        #   structure of the class and how to unflatten the flattened structure.
        #   The flatten rule for `TreeClass` is equivalent to vars(self). and the
        #   unflatten rule is equivalent to `klass(**flat_tree)`. The flatten/unflatten
        #   rule is exactly same as the flatten rule for normal dictionaries.
        treelib = sepes._src.backend.treelib
        treelib.register_treeclass(klass)

    def __setattr__(self, key: str, value: Any) -> None:
        # - implements the controlled mutability behavior.
        # - In essence, setattr is allowed to set attributes during initialization
        #   and during functional call using `value_and_tree(method)(*, **)` by marking the
        #   instnace as mutable. Otherwise, setattr is disallowed.
        # - recall that during the functional call using `value_and_tree(method)(*, **)`
        #   the tree is always copied and the copy is marked as mutable, thus
        #   setattr is allowed to set attributes on the copy not the original.
        if id(self) not in _mutable_instance_registry:
            raise AttributeError(
                f"Cannot set attribute {value=} to `{key=}`  "
                f"on an immutable instance of `{type(self).__name__}`."
            )

        getattr(object, "__setattr__")(self, key, value)

    def __delattr__(self, key: str) -> None:
        # - same as __setattr__ but for delattr.
        # - both __setattr__ and __delattr__ are used to implement the
        # - controlled mutability behavior during initialization and
        #   during functional call using `value_and_tree(method)(*, **)`.
        # - recall that during the functional call using `value_and_tree(method)(*, **)`
        #   the tree is always copied and the copy is marked as mutable, thus
        #   setattr is allowed to set attributes on the copy not the original.
        if id(self) not in _mutable_instance_registry:
            raise AttributeError(
                f"Cannot delete attribute `{key}` "
                f"on immutable instance of `{type(self).__name__}`."
            )
        getattr(object, "__delattr__")(self, key)

    @property
    def at(self) -> at[Self]:
        """Immutable out-of-place indexing.

        - ``.at[***].get()``:
            Return a new instance with the value at the index otherwise None.
        - ``.at[***].set(value)``:
            Set the `value` and return a new instance with the updated value.
        - ``.at[***].apply(func)``:
            Apply a ``func`` and return a new instance with the updated value.

        *Acceptable indexing types are:*
            - ``str`` for mapping keys or class attributes.
            - ``int`` for positional indexing for sequences.
            - ``...`` to select all leaves.
            - a boolean mask of the same structure as the tree
            - a tuple of the above types to index multiple keys at same level.

        Example:
            >>> import sepes as sp
            >>> class Tree(sp.TreeClass):
            ...    def __init__(self, a:int, b:float):
            ...        self.a = a
            ...        self.b = b
            ...    def add(self, x: int) -> int:
            ...        self.a += x
            ...        return self.a
            >>> tree = Tree(a=1, b=2.0)
            >>> tree.at["a"].get()
            Tree(a=1, b=None)
            >>> tree.at["a"].set(100)
            Tree(a=100, b=2.0)
            >>> tree.at["a"].apply(lambda x: 100)
            Tree(a=100, b=2.0)

        Note:
            - ``pytree.at[*][**]`` is equivalent to selecting pytree.*.** .
            - ``pytree.at[*, **]`` is equivalent selecting pytree.* and pytree.**
        """
        # NOTE: use `at` as a property to enable chaining syntax.
        # instead of at(at(tree)[...].apply(...))[...].set(...)
        # chaining syntax is tree.at[...].apply(...).at[...].set(...)
        return at(self)

    def __repr__(self) -> str:
        return tree_repr(self)

    def __str__(self) -> str:
        return tree_str(self)

    def __copy__(self):
        return tree_copy(self)

    def __hash__(self) -> int:
        return tree_hash(self)

    def __eq__(self, other: Any) -> bool:
        return is_tree_equal(self, other)


@tree_repr.def_type(TreeClass)
def _(node: TreeClass, **spec: Unpack[PPSpec]) -> str:
    name = type(node).__name__
    skip = [f.name for f in fields(node) if not f.repr]
    kvs = tuple((k, v) for k, v in vars(node).items() if k not in skip)
    return name + "(" + tree_repr.pps(tree_repr.av_pp, kvs, **spec) + ")"


@tree_str.def_type(TreeClass)
def _(node: TreeClass, **spec: Unpack[PPSpec]) -> str:
    name = type(node).__name__
    skip = [f.name for f in fields(node) if not f.repr]
    kvs = tuple((k, v) for k, v in vars(node).items() if k not in skip)
    return name + "(" + tree_str.pps(tree_str.av_pp, kvs, **spec) + ")"


@value_and_tree.def_mutator(TreeClass)
def _(node: TreeClass) -> None:
    add_mutable_entry(node)


@value_and_tree.def_immutator(TreeClass)
def _(node: TreeClass) -> None:
    discard_mutable_entry(node)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define lens-like indexing for pytrees

This module provides a way to index and mask pytrees (e.g. TreeClass) in an 
out-of-place manner.Out-of-place means that the original pytree is not modified, 
instead a new pytree with the selected leaves are modified.

The indexing is done through two concepts:

1) Selection (Where): Determines parts of the pytree for manipulation via a path or a boolean mask.
2) Operation (What): Defines actions on selected parts, such as setting values or applying functions.

For example, the following code defines a dict pytree with where of same structure 
as the tree. The where (Selection) defines which parts of the tree to select and 
the set (Operation) operation sets the selected parts to 100.

>>> import sepes as sp
>>> tree = {"a": 1, "b": [1, 2, 3]}
>>> where = {"a": True, "b": [False, True, False]}
>>> sp.at(tree)[where].set(100)
{'a': 100, 'b': [1, 100, 3]}
"""

from __future__ import annotations

import abc
import functools as ft
import re
from typing import Any, Callable, Generic, Hashable, TypeVar, Sequence

from typing_extensions import Self

import sepes
import sepes._src.backend.arraylib as arraylib
from sepes._src.backend import is_package_avaiable
from sepes._src.backend.treelib import ParallelConfig
from sepes._src.tree_pprint import tree_repr

T = TypeVar("T")
S = TypeVar("S")
PyTree = Any
EllipsisType = TypeVar("EllipsisType")
PathKeyEntry = TypeVar("PathKeyEntry", bound=Hashable)
_no_initializer = object()
_no_fill_value = object()


class BaseKey(abc.ABC):
    """Parent class for all match classes."""

    @abc.abstractmethod
    def __eq__(self, entry: PathKeyEntry) -> bool:
        pass

    @property
    @abc.abstractmethod
    def broadcast(self): ...


_INVALID_INDEXER = """\
Indexing with {indexer} is not implemented, supported indexing types are:
  - `str` for mapping keys or class attributes.
  - `int` for positional indexing for sequences.
  - `...` to select all leaves.
  - ``re.Pattern`` to match a leaf level path with a regex pattern.
  - Boolean mask of a compatible structure as the pytree.
  - `tuple` of the above types to match multiple leaves at the same level.
"""

_NO_LEAF_MATCH = """\
No leaf match is found for where={where}, Available keys are {names}
Check the following: 
  - If where is `str` then check if the key exists as a key or attribute.
  - If where is `int` then check if the index is in range.
  - If where is `re.Pattern` then check if the pattern matches any key.
  - If where is a `tuple` of the above types then check if any of the tuple elements match.
"""


def generate_path_mask(tree, where: tuple[BaseKey, ...], *, is_leaf=None):
    # given a pytree `tree` and a `where` path, that is composed of keys
    # generate a boolean mask that will be eventually used to with `tree_map`
    # to mark the leaves at the specified location.
    # for example for a tree = [[1, 2], 3, 4] and where = [0][1] then
    # generate [[False, True], False, False] mask
    match: bool = False
    treelib = sepes._src.backend.treelib

    def one_level_tree_path_map(func, tree):
        # apply func to the immediate children of tree
        def is_leaf_func(node) -> bool:
            # enable immediate children only
            if is_leaf and is_leaf(node) is True:
                return True
            if id(node) == id(tree):
                return False
            return True

        return treelib.path_map(func, tree, is_leaf=is_leaf_func)

    if any(where_i.broadcast for where_i in where):
        # should the selected subtree be broadcasted to the full tree
        # e.g. tree = [[1, 2], 3, 4] and where = [0], then
        # broadcast with True will be [[True, True], False, False]
        # and without broadcast will be [True, False, False]
        # the difference is that with broadcast the user defined value will
        # be broadcasted to the full subtree, for example if the user defined
        # value is 100 then the result will be [[100, 100], 3, 4]
        # and without broadcast the result will be [100, 3, 4]

        def bool_tree(value: bool, tree: Any):
            leaves, treedef = treelib.flatten(tree, is_leaf=is_leaf)
            return treelib.unflatten(treedef, [value] * len(leaves))

        true_tree = ft.partial(bool_tree, True)
        false_tree = ft.partial(bool_tree, False)

    else:
        # no broadcast, the user defined value will be applied to the selected
        # subtree only, for example if the user defined value is 100 then the
        true_tree = lambda _: True
        false_tree = lambda _: False

    def path_map_func(path, leaf):
        nonlocal match, where

        # ensure that the path is not empty
        if len(path) == len(where):
            for pi, ki in zip(path, where):
                if pi != ki:
                    return false_tree(leaf)
            match = True
            return true_tree(leaf)

        if len(path) and len(path) < len(where):
            # before traversing deeper into the tree, check if the current
            # path entry matches the current where entry, if not then return
            # a false tree to stop traversing deeper into the tree.
            (cur_where, *rest_where), (cur_path, *_) = where, path
            if cur_where == cur_path:
                # where is nonlocal to the function
                # so reduce the where path by one level and traverse deeper
                # then restore the where path to the original value before
                # returning the result
                where = rest_where
                # traverse deeper into the tree
                out_tree = one_level_tree_path_map(path_map_func, leaf)
                # return from the traversal
                where = (cur_where, *rest_where)
                return out_tree
            return false_tree(leaf)

        return false_tree(leaf)

    mask = one_level_tree_path_map(path_map_func, tree)

    if not match:
        path_leaf, _ = treelib.path_flatten(tree, is_leaf=is_leaf)
        path = "/".join(str(where_i.input) for where_i in where)
        names = "".join("\n  - " + treelib.keystr(path) for path, _ in path_leaf)
        raise LookupError(_NO_LEAF_MATCH.format(where=path, names=names))

    return mask


def resolve_where(
    where: list[Any],
    tree: T,
    is_leaf: Callable[[Any], bool] | None = None,
):
    treelib = sepes._src.backend.treelib
    ndarrays = tuple(arraylib.ndarrays)

    def combine_bool_leaves(*leaves):
        # given a list of boolean leaves, combine them using `and`
        # this is used to combine multiple boolean masks resulting from
        # either path mask or boolean mask
        verdict = True
        for leaf in leaves:
            verdict &= leaf
        return verdict

    def is_bool_leaf(leaf: Any) -> bool:
        if isinstance(leaf, ndarrays):
            return arraylib.is_bool(leaf)
        return isinstance(leaf, bool)

    # given a pytree `tree` and a `where` path, that is composed of keys or
    # boolean masks, generate a boolean mask that will be eventually used to
    # with `tree_map` to select the leaves at the specified location.
    mask = None
    bool_masks: list[T] = []
    path_masks: list[BaseKey] = []
    seen_tuple = False  # handle multiple keys at the same level
    level_paths = []

    def verify_and_aggregate_is_leaf(node: Any) -> bool:
        # use is_leaf with non-local to traverse the tree depth-first manner
        # required for verifying if a pytree is a valid indexing pytree
        nonlocal seen_tuple, level_paths, bool_masks
        # used to check if a pytree is a valid indexing pytree
        # used with `is_leaf` argument of any `tree_*` function
        leaves, _ = treelib.flatten(node)

        if all(map(is_bool_leaf, leaves)):
            # if all leaves are boolean then this is maybe a boolean mask.
            # Maybe because the boolean mask can be a valid pytree of same structure
            # as the pytree to be indexed or _compatible_ structure.
            # that can be flattend up to inside tree_map.
            # the following is an example showcase this:
            # >>> tree = [1, 2, [3, 4]]
            # >>> mask = [True, True, False]
            # >>> at(tree)[mask].get()
            # in essence the user can mark full subtrees by `False` without
            # needing to populate the subtree with `False` values. if treedef
            # check is mandated then the user will need to populate the subtree
            # with `False` values. i.e. mask = [True, True, [False, False]]
            # Finally, invalid boolean mask will be caught by `jax.tree_util`
            bool_masks += [node]
            return True

        if isinstance(resolved_key := at.dispatcher(node), BaseKey):
            # valid resolution of `BaseKey` is a valid indexing leaf
            # makes it possible to dispatch on multi-leaf pytree
            level_paths += [resolved_key]
            return False

        if type(node) is tuple and seen_tuple is False:
            # e.g. `at[1,2,3]` but not `at[1,(2,3)]``
            # i.e. inside `__getitem__` mutliple entries are transformed to a tuple
            seen_tuple = True
            return False

        # not a container of other keys or a pytree of same structure
        # emit a descriptive error message to the user by pointing to the
        # the available keys in the pytree.
        raise NotImplementedError(_INVALID_INDEXER.format(indexer=node))

    for level_keys in where:
        # each for loop iteration is a level in the where path
        # this means that if where = ("a", "b", "c") then this means
        # we are travering the tree at level "a" then level "b" then level "c"
        treelib.flatten(level_keys, is_leaf=verify_and_aggregate_is_leaf)
        # if len(level_paths) > 1 then this means that we have multiple keys
        # at the same level, for example where = ("a", ("b", "c")) then this
        # means that for a parent "a", select "b" and "c".
        path_masks += [MultiKey(*level_paths)] if len(level_paths) > 1 else level_paths
        level_paths = []
        seen_tuple = False

    if path_masks:
        mask = generate_path_mask(tree, path_masks, is_leaf=is_leaf)

    if bool_masks:
        all_masks = [mask, *bool_masks] if mask else bool_masks
        mask = treelib.map(combine_bool_leaves, *all_masks)

    return mask


class at(Generic[T]):
    """Operate on a pytree at a given path using a path or mask in out-of-place manner.

    Args:
        tree: pytree to operate on.
        where: one of the following:

            - ``str`` for mapping keys or class attributes.
            - ``int`` for positional indexing for sequences.
            - ``...`` to select all leaves.
            - a boolean mask of the same structure as the tree
            - ``re.Pattern`` to match a leaf level path with a regex pattern.
            - a tuple of the above to match multiple keys at the same level.

    Example:
        >>> import jax
        >>> import sepes as sp
        >>> tree = {"a": 1, "b": [1, 2, 3]}
        >>> sp.at(tree)["a"].set(100)
        {'a': 100, 'b': [1, 2, 3]}
        >>> sp.at(tree)["b"][0].set(100)
        {'a': 1, 'b': [100, 2, 3]}
        >>> mask = jax.tree_map(lambda x: x > 1, tree)
        >>> sp.at(tree)[mask].set(100)
        {'a': 1, 'b': [1, 100, 100]}
    """
    def __init__(self, tree: T, where: list[Any] | None = None) -> None:
        self.tree = tree
        self.where = [] if where is None else where

    def __getitem__(self, where: Any) -> Self:
        """Index a pytree at a given path using a path or mask."""
        return type(self)(self.tree, [*self.where, where])

    def __repr__(self) -> str:
        return f"{type(self).__name__}({tree_repr(self.tree)}, where={self.where})"

    def get(
        self,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
        fill_value: Any = _no_fill_value,
    ):
        """Get the leaf values at the specified location.

        Args:
            is_leaf: a predicate function to determine if a value is a leaf.
            is_parallel: accepts the following:

                - ``bool``: apply ``func`` in parallel if ``True`` otherwise in serial.
                - ``dict``: a dict of of:

                    - ``max_workers``: maximum number of workers to use.
                    - ``kind``: kind of pool to use, either ``thread`` or ``process``.

            fill_value: the value to fill the non-selected leaves with.
                Useful to use with ``jax.jit`` to avoid variable size arrays
                leaves related errors.

        Returns:
            A _new_ pytree of leaf values at the specified location, with the
            non-selected leaf values set to None if the leaf is not an array.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            >>> sp.at(tree)["b"][0].get()
            {'a': None, 'b': [1, None, None]}
        """
        treelib = sepes._src.backend.treelib
        ndarrays = tuple(arraylib.ndarrays)

        def leaf_get(where: Any, leaf: Any):
            # support both array and non-array leaves
            # for array boolean mask we select **parts** of the array that
            # matches the mask, for example if the mask is Array([True, False, False])
            # and the leaf is Array([1, 2, 3]) then the result is Array([1])
            # because of the variable resultant size of the output
            if isinstance(where, ndarrays) and len(arraylib.shape(where)):
                if fill_value is not _no_fill_value:
                    return arraylib.where(where, leaf, fill_value)
                return leaf[where]
            # non-array boolean mask we select the leaf if the mask is True
            # and `None` otherwise
            if fill_value is not _no_fill_value:
                return leaf if where else fill_value
            return leaf if where else None

        return treelib.map(
            leaf_get,
            resolve_where(self.where, self.tree, is_leaf),
            self.tree,
            is_leaf=is_leaf,
            is_parallel=is_parallel,
        )

    def set(
        self,
        set_value: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ):
        """Set the leaf values at the specified location.

        Args:
            set_value: the value to set at the specified location.
            is_leaf: a predicate function to determine if a value is a leaf.
            is_parallel: accepts the following:

                - ``bool``: apply ``func`` in parallel if ``True`` otherwise in serial.
                - ``dict``: a dict of of:

                    - ``max_workers``: maximum number of workers to use.
                    - ``kind``: kind of pool to use, either ``thread`` or ``process``.

        Returns:
            A pytree with the leaf values at the specified location
            set to ``set_value``.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            >>> sp.at(tree)["b"][0].set(100)
            {'a': 1, 'b': [100, 2, 3]}
        """
        treelib = sepes._src.backend.treelib
        ndarrays = tuple(arraylib.ndarrays)

        def leaf_set(where: Any, leaf: Any, set_value: Any):
            # support both array and non-array leaves
            # for array boolean mask we select **parts** of the array that
            # matches the mask, for example if the mask is Array([True, False, False])
            # and the leaf is Array([1, 2, 3]) then the result is Array([1, 100, 100])
            # with set_value = 100
            if isinstance(where, ndarrays):
                return arraylib.where(where, set_value, leaf)
            return set_value if where else leaf

        _, lhsdef = treelib.flatten(self.tree, is_leaf=is_leaf)
        _, rhsdef = treelib.flatten(set_value, is_leaf=is_leaf)

        if lhsdef == rhsdef:
            # do not broadcast set_value if it is a pytree of same structure
            # for example tree.at[where].set(tree2) will set all tree leaves
            # to tree2 leaves if tree2 is a pytree of same structure as tree
            # instead of making each leaf of tree a copy of tree2
            # is design is similar to ``numpy`` design `np.at[...].set(Array)`
            return treelib.map(
                leaf_set,
                resolve_where(self.where, self.tree, is_leaf),
                self.tree,
                set_value,
                is_leaf=is_leaf,
                is_parallel=is_parallel,
            )

        return treelib.map(
            ft.partial(leaf_set, set_value=set_value),
            resolve_where(self.where, self.tree, is_leaf),
            self.tree,
            is_leaf=is_leaf,
            is_parallel=is_parallel,
        )

    def apply(
        self,
        func: Callable[[Any], Any],
        *,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ):
        """Apply a function to the leaf values at the specified location.

        Args:
            func: the function to apply to the leaf values.
            is_leaf: a predicate function to determine if a value is a leaf.
            is_parallel: accepts the following:

                - ``bool``: apply ``func`` in parallel if ``True`` otherwise in serial.
                - ``dict``: a dict of of:

                    - ``max_workers``: maximum number of workers to use.
                    - ``kind``: kind of pool to use, either ``thread`` or ``process``.

        Returns:
            A pytree with the leaf values at the specified location set to
            the result of applying ``func`` to the leaf values.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            >>> sp.at(tree)["b"][0].apply(lambda x: x + 100)
            {'a': 1, 'b': [101, 2, 3]}

        Example:
            Read images in parallel

            >>> import sepes as sp
            >>> from matplotlib.pyplot import imread
            >>> path = {"img1": "path1.png", "img2": "path2.png"}
            >>> is_parallel = dict(max_workers=2)
            >>> images = sp.at(path)[...].apply(imread, is_parallel=is_parallel)  # doctest: +SKIP
        """
        treelib = sepes._src.backend.treelib
        ndarrays = tuple(arraylib.ndarrays)

        def leaf_apply(where: Any, leaf: Any):
            # same as `leaf_set` but with `func` applied to the leaf
            # one thing to note is that, the where mask select an array
            # then the function needs work properly when applied to the selected
            # array elements
            if isinstance(where, ndarrays):
                return arraylib.where(where, func(leaf), leaf)
            return func(leaf) if where else leaf

        return treelib.map(
            leaf_apply,
            resolve_where(self.where, self.tree, is_leaf),
            self.tree,
            is_leaf=is_leaf,
            is_parallel=is_parallel,
        )

    def scan(
        self,
        func: Callable[[Any, S], tuple[Any, S]],
        state: S,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Any, S]:
        """Apply a function while carrying a state.

        Args:
            func: the function to apply to the leaf values. the function accepts
                a running state and leaf value and returns a tuple of the new
                leaf value and the new state.
            state: the initial state to carry.
            is_leaf: a predicate function to determine if a value is a leaf. for
                example, ``lambda x: isinstance(x, list)`` will treat all lists
                as leaves and will not recurse into list items.

        Returns:
            A tuple of the final state and pytree with the leaf values at the
            specified location set to the result of applying ``func`` to the leaf
            values.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            >>> def scan_func(leaf, running_max):
            ...     cur_max = max(leaf, running_max)
            ...     return leaf, cur_max
            >>> running_max = float("-inf")
            >>> _, running_max = sp.at(tree)["b"][0, 1].scan(scan_func, state=running_max)
            >>> running_max  # max of b[0] and b[1]
            2

        Note:
            ``scan`` applies a binary ``func`` to the leaf values while carrying
            a state and returning a tree leaves with the the ``func`` applied to
            them with final state. While ``reduce`` applies a binary ``func`` to the
            leaf values while carrying a state and returning a single value.
        """
        treelib = sepes._src.backend.treelib
        ndarrays = tuple(arraylib.ndarrays)
        running_state = state

        def stateless_func(leaf):
            nonlocal running_state
            leaf, running_state = func(leaf, running_state)
            return leaf

        def leaf_apply(where: Any, leaf: Any):
            if isinstance(where, ndarrays):
                return arraylib.where(where, stateless_func(leaf), leaf)
            return stateless_func(leaf) if where else leaf

        out_tree = treelib.map(
            leaf_apply,
            resolve_where(self.where, self.tree, is_leaf),
            self.tree,
            is_leaf=is_leaf,
        )
        return out_tree, running_state

    def reduce(
        self,
        func: Callable[[Any, Any], Any],
        *,
        initializer: Any = _no_initializer,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> Any:
        """Reduce the leaf values at the specified location.

        Args:
            func: the function to reduce the leaf values.
            initializer: the initializer value for the reduction.
            is_leaf: a predicate function to determine if a value is a leaf.

        Returns:
            The result of reducing the leaf values at the specified location.

        Note:
            - If ``initializer`` is not specified, the first leaf value is used as
              the initializer.
            - ``reduce`` applies a binary ``func`` to each leaf values while accumulating
              a state a returns the final result. while ``scan`` applies ``func`` to each
              leaf value while carrying a state and returns the final state and
              the leaves of the tree with the result of applying ``func`` to each leaf.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            >>> sp.at(tree)["b"].reduce(lambda x, y: x + y)
            6
        """
        treelib = sepes._src.backend.treelib
        tree = self.get(is_leaf=is_leaf)  # type: ignore
        leaves, _ = treelib.flatten(tree, is_leaf=is_leaf)
        if initializer is _no_initializer:
            return ft.reduce(func, leaves)
        return ft.reduce(func, leaves, initializer)

    def pluck(
        self,
        count: int | None = None,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> list[Any]:
        """Extract subtrees at the specified location.

        Note:
            ``pluck`` first applies ``get`` to the specified location and then
            extracts the immediate subtrees of the selected leaves. ``is_leaf``
            and ``is_parallel`` are passed to ``get``.

        Args:
            count: number of subtrees to extract, Default to ``None`` to
                extract all subtrees.
            is_leaf: a predicate function to determine if a value is a leaf.
            is_parallel: accepts the following:

                - ``bool``: apply ``func`` in parallel if ``True`` otherwise in serial.
                - ``dict``: a dict of of:

                    - ``max_workers``: maximum number of workers to use.
                    - ``kind``: kind of pool to use, either ``thread`` or ``process``.

        Returns:
            A list of subtrees at the specified location.

        Note:
            Compared to ``get``, ``pluck`` extracts subtrees at the specified
            location and returns a list of subtrees. While ``get`` returns a
            pytree with the leaf values at the specified location and set the
            non-selected leaf values to ``None``.

        Example:
            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [1, 2, 3]}
            <BLANKLINE>
            >>> # `pluck` returns a list of selected subtrees
            >>> sp.at(tree)["b"].pluck()
            [[1, 2, 3]]
            <BLANKLINE>
            >>> # `get` returns same pytree
            >>> sp.at(tree)["b"].get()
            {'a': None, 'b': [1, 2, 3]}

        Example:
            ``pluck`` with mask

            >>> import sepes as sp
            >>> tree = {"a": 1, "b": [2, 3, 4]}
            >>> mask = {"a": True, "b": [False, True, False]}
            >>> sp.at(tree)[mask].pluck()
            [1, 3]

            This is equivalent to the following:

            >>> [tree["a"], tree["b"][1]]  # doctest: +SKIP
        """
        treelib = sepes._src.backend.treelib
        tree = self.get(is_leaf=is_leaf, is_parallel=is_parallel)
        subtrees: list[Any] = []
        count = float("inf") if count is None else count

        def aggregate_subtrees(node: Any) -> bool:
            nonlocal subtrees, count
            if count < 1:
                # stop traversing the tree
                # if total number of subtrees is reached
                return True
            if id(node) == id(tree):
                # skip the root node
                # for example if tree = dict(a=1) and mask is dict(a=True)
                # then returns [1] and not [dict(a=1)]
                return False
            leaves, _ = treelib.flatten(node, is_leaf=lambda x: x is None)
            # in essence if the subtree does not contain any None leaves
            # then it is a valid subtree to be plucked
            # this because `get` sets the non-selected leaves to None
            if any(leaf is None for leaf in leaves):
                return False
            subtrees += [node]
            count -= 1
            return True

        treelib.flatten(tree, is_leaf=aggregate_subtrees)
        return subtrees


# pass through for boolean pytrees masks and tuple of keys
at.dispatcher = ft.singledispatch(lambda x: x)


def def_rule(
    user_type: type[T],
    path_compare_func: Callable[[T, PathKeyEntry], bool],
    *,
    broadcastable: bool = False,
) -> None:
    # remove the BaseKey abstraction from the user-facing function
    class UserKey(BaseKey):
        broadcast: bool = broadcastable

        def __init__(self, input: T):
            self.input = input

        def __eq__(self, key: PathKeyEntry) -> bool:
            return path_compare_func(self.input, key)

    at.dispatcher.register(user_type, UserKey)


at.def_rule = def_rule


# key rules to match user input to with the path entry


def str_compare(name: str, key: PathKeyEntry):
    """Match a leaf with a given name."""
    if isinstance(key, str):
        return name == key
    treelib = sepes._src.backend.treelib
    if isinstance(key, type(treelib.attribute_key(""))):
        return name == key.name
    if isinstance(key, type(treelib.dict_key(""))):
        return name == key.key
    return False


def int_compare(idx: int, key: PathKeyEntry) -> bool:
    """Match a leaf with a given index."""
    if isinstance(key, int):
        return idx == key
    treelib = sepes._src.backend.treelib
    if isinstance(key, type(treelib.sequence_key(0))):
        return idx == key.idx
    return False


def regex_compare(pattern: re.Pattern, key: PathKeyEntry) -> bool:
    """Match a path with a regex pattern inside 'at' property."""
    if isinstance(key, str):
        return re.fullmatch(pattern, key) is not None
    treelib = sepes._src.backend.treelib
    if isinstance(key, type(treelib.attribute_key(""))):
        return re.fullmatch(pattern, key.name) is not None
    if isinstance(key, type(treelib.dict_key(""))):
        return re.fullmatch(pattern, key.key) is not None
    return False


def ellipsis_compare(_, __):
    return True


at.def_rule(str, str_compare, broadcastable=False)
at.def_rule(int, int_compare, broadcastable=False)
at.def_rule(re.Pattern, regex_compare, broadcastable=False)
at.def_rule(type(...), ellipsis_compare, broadcastable=True)


class MultiKey(BaseKey):
    """Match a leaf with multiple keys at the same level."""

    def __init__(self, *keys):
        self.keys = tuple(keys)

    def __eq__(self, entry: PathKeyEntry) -> bool:
        return any(entry == key for key in self.keys)

    broadcast: bool = False


if is_package_avaiable("jax"):
    import jax.tree_util as jtu

    def jax_key_compare(input, key: PathKeyEntry) -> bool:
        """Enable indexing with jax keys directly in `at`."""
        return input == key

    at.def_rule(jtu.SequenceKey, jax_key_compare, broadcastable=False)
    at.def_rule(jtu.GetAttrKey, jax_key_compare, broadcastable=False)
    at.def_rule(jtu.DictKey, jax_key_compare, broadcastable=False)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sepes._src.backend import backend_context
from sepes._src.code_build import autoinit, field, fields
from sepes._src.tree_base import TreeClass
from sepes._src.tree_index import at
from sepes._src.tree_mask import is_masked, tree_mask, tree_unmask
from sepes._src.tree_pprint import tree_diagram, tree_repr, tree_str, tree_summary
from sepes._src.tree_util import bcmap, leafwise, value_and_tree

__all__ = [
    # module utils
    "TreeClass",
    # pprint utils
    "tree_diagram",
    "tree_repr",
    "tree_str",
    "tree_summary",
    # masking utils
    "is_masked",
    "tree_unmask",
    "tree_mask",
    # tree utils
    "at",
    "bcmap",
    "value_and_tree",
    # construction utils
    "field",
    "fields",
    "autoinit",
    "leafwise",
    # backend utils
    "backend_context",
]

__version__ = "0.12.0"

at.__module__ = "sepes"
TreeClass.__module__ = "sepes"



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses as dc
from typing import Any, Callable, Hashable, Iterable

import jax.tree_util as jtu

from sepes._src.backend.treelib import (
    AbstractTreeLib,
    KeyPathLeaf,
    ParallelConfig,
    Tree,
    concurrent_map,
)


@dc.dataclass(frozen=True)
class NamedSequenceKey(jtu.GetAttrKey, jtu.SequenceKey):
    def __str__(self):
        return f".{self.name}"


class JaxTreeLib(AbstractTreeLib):
    @staticmethod
    def map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        leaves, treedef = jtu.tree_flatten(tree, is_leaf)
        flat: list[Any] = [leaves] + [treedef.flatten_up_to(r) for r in rest]
        if is_parallel is False:
            return jtu.tree_unflatten(treedef, [func(*args) for args in zip(*flat)])
        config = dict() if is_parallel is True else is_parallel
        return jtu.tree_unflatten(treedef, concurrent_map(func, flat, **config))

    @staticmethod
    def path_map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        leaves, treedef = jtu.tree_flatten_with_path(tree, is_leaf)
        flat: list[Any] = list(zip(*leaves)) + [treedef.flatten_up_to(r) for r in rest]
        if not is_parallel:
            return jtu.tree_unflatten(treedef, [func(*args) for args in zip(*flat)])
        config = dict() if is_parallel is True else is_parallel
        return jtu.tree_unflatten(treedef, concurrent_map(func, flat, **config))

    @staticmethod
    def flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[Any], jtu.PyTreeDef]:
        return jtu.tree_flatten(tree, is_leaf=is_leaf)

    @staticmethod
    def path_flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[KeyPathLeaf], jtu.PyTreeDef]:
        return jtu.tree_flatten_with_path(tree, is_leaf=is_leaf)

    @staticmethod
    def unflatten(treedef: jtu.PyTreeDef, leaves: Iterable[Any]) -> Any:
        return jtu.tree_unflatten(treedef, leaves)

    @staticmethod
    def register_treeclass(klass: type[Tree]) -> None:
        def unflatten(keys: tuple[str, ...], leaves: tuple[Any, ...]) -> Tree:
            vars(tree := getattr(object, "__new__")(klass)).update(zip(keys, leaves))
            return tree

        def flatten(tree: Tree) -> tuple[tuple[Any, ...], tuple[str, ...]]:
            return tuple((dynamic := vars(tree)).values()), tuple(dynamic.keys())

        def flatten_with_keys(tree: Tree):
            dynamic = dict(vars(tree))
            for idx, key in enumerate(vars(tree)):
                dynamic[key] = (NamedSequenceKey(idx, key), dynamic[key])
            return tuple(dynamic.values()), tuple(dynamic.keys())

        jtu.register_pytree_with_keys(klass, flatten_with_keys, unflatten, flatten)

    @staticmethod
    def register_static(klass: type[Tree]) -> None:
        jtu.register_pytree_node(klass, lambda x: ((), x), lambda x, _: x)

    @staticmethod
    def attribute_key(name: str) -> jtu.GetAttrKey:
        return jtu.GetAttrKey(name)

    @staticmethod
    def sequence_key(index: int) -> jtu.SequenceKey:
        return jtu.SequenceKey(index)

    @staticmethod
    def dict_key(key: Hashable) -> jtu.DictKey:
        return jtu.DictKey(key)

    @staticmethod
    def keystr(keys: Any) -> str:
        return jtu.keystr(keys)
